from .utils import clean_setting


DISCORD_APP_ID = clean_setting('DISCORD_APP_ID', '')
"""App ID for the AA bot on Discord. Needs to be set."""

DISCORD_APP_SECRET = clean_setting('DISCORD_APP_SECRET', '')
"""App secret for the AA bot on Discord. Needs to be set."""

DISCORD_BOT_TOKEN = clean_setting('DISCORD_BOT_TOKEN', '')
"""Token used by the AA bot on Discord.  Needs to be set."""

DISCORD_CALLBACK_URL = clean_setting('DISCORD_CALLBACK_URL', '')
"""Callback URL for OAuth with Discord. Needs to be set."""

DISCORD_GUILD_ID = clean_setting('DISCORD_GUILD_ID', '')
"""ID of the Discord Server. Needs to be set."""

DISCORD_TASKS_MAX_RETRIES = clean_setting('DISCORD_TASKS_MAX_RETRIES', 3)
"""Max retries of tasks after an error occurred."""

DISCORD_TASKS_RETRY_PAUSE = clean_setting('DISCORD_TASKS_RETRY_PAUSE', 60)
"""Pause in seconds until next retry for tasks after the API returned an error."""

DISCORD_SYNC_NAMES = clean_setting('DISCORD_SYNC_NAMES', False)
"""Automatically sync Discord users names to user's main character name when created."""



import logging

from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.contrib.auth.decorators import permission_required
from django.shortcuts import get_object_or_404
from django.shortcuts import render, redirect
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from .form import OpForm
from .models import OpTimer, OpTimerType

logger = logging.getLogger(__name__)


@login_required
@permission_required('auth.optimer_view')
def optimer_view(request):
    logger.debug("optimer_view called by user %s" % request.user)
    base_query = OpTimer.objects.select_related('eve_character', 'type')
    render_items = {'optimer': base_query.all(),
                    'future_timers': base_query.filter(
                        start__gte=timezone.now()),
                    'past_timers': base_query.filter(
                        start__lt=timezone.now()).order_by('-start')}

    return render(request, 'optimer/management.html', context=render_items)


@login_required
@permission_required('auth.optimer_management')
def add_optimer_view(request):
    logger.debug("add_optimer_view called by user %s" % request.user)
    if request.method == 'POST':
        form = OpForm(request.POST, data_list=OpTimerType.objects.all())
        logger.debug("Request type POST contains form valid: %s" % form.is_valid())
        if form.is_valid():
            optimer_type = None

            if form.cleaned_data['type'] != '':
                try:
                    optimer_type = OpTimerType.objects.get(
                        type__iexact=form.cleaned_data['type']
                    )
                except OpTimerType.DoesNotExist:
                    optimer_type = OpTimerType.objects.create(
                        type=form.cleaned_data['type']
                    )

            # Get Current Time
            post_time = timezone.now()
            # Get character
            character = request.user.profile.main_character
            # handle valid form
            op = OpTimer()
            op.doctrine = form.cleaned_data['doctrine']
            op.system = form.cleaned_data['system']
            op.start = form.cleaned_data['start']
            op.duration = form.cleaned_data['duration']
            op.operation_name = form.cleaned_data['operation_name']
            op.fc = form.cleaned_data['fc']
            op.create_time = post_time
            op.eve_character = character
            op.type = optimer_type
            op.description = form.cleaned_data['description']
            op.save()
            logger.info(f"User {request.user} created op timer with name {op.operation_name}")
            messages.success(request, _('Created operation timer for %(opname)s.') % {"opname": op.operation_name})
            return redirect("optimer:view")
    else:
        logger.debug("Returning new opForm")
        form = OpForm(data_list=OpTimerType.objects.all())

    render_items = {'form': form}

    return render(request, 'optimer/add.html', context=render_items)


@login_required
@permission_required('auth.optimer_management')
def remove_optimer(request, optimer_id):
    logger.debug(f"remove_optimer called by user {request.user} for operation id {optimer_id}")
    op = get_object_or_404(OpTimer, id=optimer_id)
    op.delete()
    logger.info(f"Deleting optimer id {optimer_id} by user {request.user}")
    messages.success(request, _('Removed operation timer for %(opname)s.') % {"opname": op.operation_name})
    return redirect("optimer:view")


@login_required
@permission_required('auth.optimer_management')
def edit_optimer(request, optimer_id):
    logger.debug(f"edit_optimer called by user {request.user} for optimer id {optimer_id}")
    op = get_object_or_404(OpTimer, id=optimer_id)
    if request.method == 'POST':
        form = OpForm(request.POST, data_list=OpTimerType.objects.all())
        logger.debug("Received POST request containing update optimer form, is valid: %s" % form.is_valid())
        if form.is_valid():
            character = request.user.profile.main_character

            optimer_type = None

            if form.cleaned_data['type'] != '':
                try:
                    optimer_type = OpTimerType.objects.get(
                        type__iexact=form.cleaned_data['type']
                    )
                except OpTimerType.DoesNotExist:
                    optimer_type = OpTimerType.objects.create(
                        type=form.cleaned_data['type']
                    )

            op.doctrine = form.cleaned_data['doctrine']
            op.system = form.cleaned_data['system']
            op.start = form.cleaned_data['start']
            op.duration = form.cleaned_data['duration']
            op.operation_name = form.cleaned_data['operation_name']
            op.fc = form.cleaned_data['fc']
            op.eve_character = character
            op.type = optimer_type
            op.description = form.cleaned_data['description']
            logger.info(f"User {request.user} updating optimer id {optimer_id} ")
            op.save()
            messages.success(request, _('Saved changes to operation timer for %(opname)s.') % {"opname": op.operation_name})
            return redirect("optimer:view")
    else:
        data = {
            'doctrine': op.doctrine,
            'system': op.system,
            'start': op.start,
            'duration': op.duration,
            'operation_name': op.operation_name,
            'fc': op.fc,
            'description': op.description,
            'type': op.type
        }
        form = OpForm(initial=data, data_list=OpTimerType.objects.all())
    return render(request, 'optimer/update.html', context={'form': form})


def dashboard_ops(request):
    base_query = OpTimer.objects.select_related('eve_character', 'type')
    timers = base_query.filter(start__gte=timezone.now())[:5]
    if timers.count():
        context = {
            'timers': timers,
        }
        return render_to_string('optimer/dashboard.ops.html', context=context, request=request)
    else:
        return ""



# Generated by Django 4.0.2 on 2022-08-28 14:00

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='MenuItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('hook_function', models.CharField(max_length=500)),
                ('icon_classes', models.CharField(max_length=150)),
                ('text', models.CharField(max_length=150)),
                ('url', models.CharField(blank=True, default=None, max_length=2048, null=True)),
                ('rank', models.IntegerField(default=1000)),
                ('hide', models.BooleanField(default=False)),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='menu.menuitem')),
            ],
        ),
    ]



from unittest import mock

from django.contrib.auth.models import Group
from django.test import TestCase, override_settings

from allianceauth.tests.auth_utils import AuthUtils

from ..models import GroupRequest, RequestLog, ReservedGroupName

MODULE_PATH = "allianceauth.groupmanagement.models"


def create_testdata():
    # group 1
    group = Group.objects.create(name='Superheros')
    group.authgroup.description = 'Default Group'
    group.authgroup.internal = False
    group.authgroup.hidden = False
    group.authgroup.save()
    # user 1
    user_1 = AuthUtils.create_user('Bruce Wayne')
    AuthUtils.add_main_character_2(
        user_1,
        name='Bruce Wayne',
        character_id=1001,
        corp_id=2001,
        corp_name='Wayne Technologies'
    )
    user_1.groups.add(group)
    group.authgroup.group_leaders.add(user_1)
    # user 2
    user_2 = AuthUtils.create_user('Clark Kent')
    AuthUtils.add_main_character_2(
        user_2,
        name='Clark Kent',
        character_id=1002,
        corp_id=2002,
        corp_name='Wayne Food'
    )
    # user 3
    user_3 = AuthUtils.create_user('Peter Parker')
    AuthUtils.add_main_character_2(
        user_2,
        name='Peter Parker',
        character_id=1003,
        corp_id=2002,
        corp_name='Wayne Food'
    )
    return group, user_1, user_2, user_3


class TestGroupRequest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.group, cls.user_1, cls.user_2, cls.user_3 = create_testdata()

    def test_main_char(self):
        group_request = GroupRequest.objects.create(
            user=self.user_1,
            group=self.group
        )
        expected = self.user_1.profile.main_character
        self.assertEqual(group_request.main_char, expected)

    def test_str(self):
        group_request = GroupRequest.objects.create(
            user=self.user_1,
            group=self.group
        )
        expected = 'Bruce Wayne:Superheros'
        self.assertEqual(str(group_request), expected)

    @override_settings(GROUPMANAGEMENT_REQUESTS_NOTIFICATION=True)
    def test_should_notify_leaders_about_join_request(self):
        # given
        group_request = GroupRequest.objects.create(
            user=self.user_2, group=self.group
        )
        # when
        with mock.patch(MODULE_PATH + ".notify") as mock_notify:
            group_request.notify_leaders()
        # then
        self.assertTrue(mock_notify.called)
        _, kwargs = mock_notify.call_args
        self.assertEqual(kwargs["user"],self.user_1)

    @override_settings(GROUPMANAGEMENT_REQUESTS_NOTIFICATION=True)
    def test_should_notify_leaders_about_leave_request(self):
        # given
        group_request = GroupRequest.objects.create(
            user=self.user_2, group=self.group
        )
        # when
        with mock.patch(MODULE_PATH + ".notify") as mock_notify:
            group_request.notify_leaders()
        # then
        self.assertTrue(mock_notify.called)

    @override_settings(GROUPMANAGEMENT_REQUESTS_NOTIFICATION=True)
    def test_should_handle_notify_leaders_without_leaders(self):
        # given
        group = Group.objects.create(name='Dummy')
        group.authgroup.internal = False
        group.authgroup.hidden = False
        group.authgroup.save()
        group_request = GroupRequest.objects.create(
            user=self.user_2, group=group
        )
        # when
        with mock.patch(MODULE_PATH + ".notify") as mock_notify:
            group_request.notify_leaders()
        # then
        self.assertFalse(mock_notify.called)

    @override_settings(GROUPMANAGEMENT_REQUESTS_NOTIFICATION=False)
    def test_should_not_notify_leaders_if_disabled(self):
        # given
        group_request = GroupRequest.objects.create(
            user=self.user_2, group=self.group
        )
        # when
        with mock.patch(MODULE_PATH + ".notify") as mock_notify:
            group_request.notify_leaders()
        # then
        self.assertFalse(mock_notify.called)

    @override_settings(GROUPMANAGEMENT_REQUESTS_NOTIFICATION=True)
    def test_should_notify_members_of_leader_groups_about_join_request(self):
        # given
        child_group = Group.objects.create(name='Child')
        child_group.authgroup.internal = False
        child_group.authgroup.hidden = False
        child_group.authgroup.save()
        child_group.authgroup.group_leader_groups.add(self.group)
        group_request = GroupRequest.objects.create(
            user=self.user_2, group=child_group
        )
        # when
        with mock.patch(MODULE_PATH + ".notify") as mock_notify:
            group_request.notify_leaders()
        # then
        self.assertTrue(mock_notify.called)
        _, kwargs = mock_notify.call_args
        self.assertEqual(kwargs["user"],self.user_1)


class TestRequestLog(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.group, cls.user_1, cls.user_2, _ = create_testdata()

    def test_requestor(self):
        request_log = RequestLog.objects.create(
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1
        )
        expected = 'Clark Kent'
        self.assertEqual(request_log.requestor(), expected)

    def test_type_to_str_removed(self):
        request_log = RequestLog.objects.create(
            request_type=None,
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1
        )
        expected = 'Removed'
        self.assertEqual(request_log.type_to_str(), expected)

    def test_type_to_str_leave(self):
        request_log = RequestLog.objects.create(
            request_type=True,
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1
        )
        expected = 'Leave'
        self.assertEqual(request_log.type_to_str(), expected)

    def test_type_to_str_join(self):
        request_log = RequestLog.objects.create(
            request_type=False,
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1
        )
        expected = 'Join'
        self.assertEqual(request_log.type_to_str(), expected)

    def test_action_to_str_accept(self):
        request_log = RequestLog.objects.create(
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1,
            action=True
        )
        expected = 'Accept'
        self.assertEqual(request_log.action_to_str(), expected)

    def test_action_to_str_reject(self):
        request_log = RequestLog.objects.create(
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1,
            action=False
        )
        expected = 'Reject'
        self.assertEqual(request_log.action_to_str(), expected)

    def test_req_char(self):
        request_log = RequestLog.objects.create(
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1,
            action=False
        )
        expected = self.user_2.profile.main_character
        self.assertEqual(request_log.req_char(), expected)


class TestAuthGroup(TestCase):
    def test_str(self):
        group = Group.objects.create(name='Superheros')
        group.authgroup.description = 'Default Group'
        group.authgroup.internal = False
        group.authgroup.hidden = False
        group.authgroup.save()

        expected = 'Superheros'
        self.assertEqual(str(group.authgroup), expected)

    def test_should_remove_guests_from_group_when_restricted_to_members_only(self):
        # given
        user_member = AuthUtils.create_user("Bruce Wayne")
        character_member = AuthUtils.add_main_character_2(
            user_member,
            name="Bruce Wayne",
            character_id=1001,
            corp_id=2001,
            corp_name="Wayne Technologies",
        )
        user_guest = AuthUtils.create_user("Lex Luthor")
        AuthUtils.add_main_character_2(
            user_guest,
            name="Lex Luthor",
            character_id=1011,
            corp_id=2011,
            corp_name="Luthor Corp",
        )
        member_state = AuthUtils.get_member_state()
        member_state.member_characters.add(character_member)
        user_member.refresh_from_db()
        user_guest.refresh_from_db()
        group = Group.objects.create(name="dummy")
        user_member.groups.add(group)
        user_guest.groups.add(group)
        group.authgroup.states.add(member_state)
        # when
        group.authgroup.remove_users_not_matching_states()
        # then
        self.assertIn(group, user_member.groups.all())
        self.assertNotIn(group, user_guest.groups.all())


class TestAuthGroupRequestApprovers(TestCase):
    def setUp(self) -> None:
        self.group, self.user_1, self.user_2, self.user_3 = create_testdata()

    def test_should_return_leaders_of_main_group_only(self):
        # when
        leaders = self.group.authgroup.group_request_approvers()
        # then
        self.assertSetEqual(leaders, {self.user_1})

    def test_should_return_members_of_leading_groups_only(self):
        # given
        parent_group = Group.objects.create(name='Parent')
        parent_group.authgroup.group_leaders.add(self.user_2)
        self.user_1.groups.add(parent_group)
        child_group = Group.objects.create(name='Child')
        child_group.authgroup.internal = False
        child_group.authgroup.hidden = False
        child_group.authgroup.save()
        child_group.authgroup.group_leader_groups.add(parent_group)
        # when
        leaders = child_group.authgroup.group_request_approvers()
        # then
        self.assertSetEqual(leaders, {self.user_1})

    def test_should_return_leaders_of_main_group_and_members_of_leading_groups(self):
        # given
        parent_group = Group.objects.create(name='Parent')
        parent_group.authgroup.group_leaders.add(self.user_2)
        self.user_1.groups.add(parent_group)
        child_group = Group.objects.create(name='Child')
        child_group.authgroup.internal = False
        child_group.authgroup.hidden = False
        child_group.authgroup.save()
        child_group.authgroup.group_leaders.add(self.user_3)
        child_group.authgroup.group_leader_groups.add(self.group)
        # when
        leaders = child_group.authgroup.group_request_approvers()
        # then
        self.assertSetEqual(leaders, {self.user_1, self.user_3})

    def test_can_handle_group_without_leaders(self):
        # given
        child_group = Group.objects.create(name='Child')
        child_group.authgroup.internal = False
        child_group.authgroup.hidden = False
        child_group.authgroup.save()
        # when
        leaders = child_group.authgroup.group_request_approvers()
        # then
        self.assertSetEqual(leaders, set())


class TestReservedGroupName(TestCase):
    def test_should_return_name(self):
        # given
        obj = ReservedGroupName(name="test", reason="abc", created_by="xxx")
        # when
        result = str(obj)
        # then
        self.assertEqual(result, "test")

    def test_should_not_allow_creating_reserved_name_for_existing_group(self):
        # given
        Group.objects.create(name="Dummy")
        # when
        with self.assertRaises(RuntimeError):
            ReservedGroupName.objects.create(
                name="dummy", reason="abc", created_by="xxx"
            )



import logging

from .models import (
    CharacterOwnership,
    UserProfile,
    get_guest_state,
    State,
    OwnershipRecord)
from django.contrib.auth.models import User
from django.db.models import Q
from django.db.models.signals import pre_save, post_save, pre_delete, post_delete, m2m_changed
from django.dispatch import receiver, Signal
from esi.models import Token

from allianceauth.eveonline.models import EveCharacter

logger = logging.getLogger(__name__)

state_changed = Signal()


def trigger_state_check(state):
    # evaluate all current members to ensure they still have access
    for profile in state.userprofile_set.all():
        profile.assign_state()

    # we may now be available to others with lower states
    check_states = State.objects.filter(priority__lt=state.priority)
    for profile in UserProfile.objects.filter(state__in=check_states):
        if state.available_to_user(profile.user):
            profile.assign_state(state)


@receiver(m2m_changed, sender=State.member_characters.through)
def state_member_characters_changed(sender, instance, action, *args, **kwargs):
    if action.startswith('post_'):
        logger.debug(f'State {instance} member characters changed. Re-evaluating membership.')
        trigger_state_check(instance)


@receiver(m2m_changed, sender=State.member_corporations.through)
def state_member_corporations_changed(sender, instance, action, *args, **kwargs):
    if action.startswith('post_'):
        logger.debug(f'State {instance} member corporations changed. Re-evaluating membership.')
        trigger_state_check(instance)


@receiver(m2m_changed, sender=State.member_alliances.through)
def state_member_alliances_changed(sender, instance, action, *args, **kwargs):
    if action.startswith('post_'):
        logger.debug(f'State {instance} member alliances changed. Re-evaluating membership.')
        trigger_state_check(instance)

@receiver(m2m_changed, sender=State.member_factions.through)
def state_member_factions_changed(sender, instance, action, *args, **kwargs):
    if action.startswith('post_'):
        logger.debug(f'State {instance} member factions changed. Re-evaluating membership.')
        trigger_state_check(instance)

@receiver(post_save, sender=State)
def state_saved(sender, instance, *args, **kwargs):
    logger.debug(f'State {instance} saved. Re-evaluating membership.')
    trigger_state_check(instance)


# Is there a smarter way to intercept pre_save with a diff main_character or state?
@receiver(post_save, sender=UserProfile)
def reassess_on_profile_save(sender, instance, created, *args, **kwargs):
    # catches post_save from profiles to trigger necessary service and state checks
    if not created:
        update_fields = kwargs.pop('update_fields', []) or []
        if 'state' not in update_fields:
            logger.debug(f'Profile for {instance.user} saved without state change. Re-evaluating state.')
            instance.assign_state()


@receiver(post_save, sender=User)
def create_required_models(sender, instance, created, *args, **kwargs):
    # ensure all users have our Sub-Models
    if created:
        logger.debug(f'User {instance} created. Creating default UserProfile.')
        UserProfile.objects.get_or_create(user=instance)


@receiver(post_save, sender=Token)
def record_character_ownership(sender, instance, created, *args, **kwargs):
    if created:
        logger.debug(f'New token for {instance.user} character {instance.character_name} saved. Evaluating ownership.')
        if instance.user:
            query = Q(owner_hash=instance.character_owner_hash) & Q(user=instance.user)
        else:
            query = Q(owner_hash=instance.character_owner_hash)
        # purge ownership records if the hash or auth user account has changed
        CharacterOwnership.objects.filter(character__character_id=instance.character_id).exclude(query).delete()
        # create character if needed
        if EveCharacter.objects.filter(character_id=instance.character_id).exists() is False:
            logger.debug(f'Token is for a new character. Creating model for {instance.character_name} ({instance.character_id})')
            EveCharacter.objects.create_character(instance.character_id)
        char = EveCharacter.objects.get(character_id=instance.character_id)
        # check if we need to create ownership
        if instance.user and not CharacterOwnership.objects.filter(
                character__character_id=instance.character_id).exists():
            logger.debug(f"Character {instance.character_name} is not yet owned. Assigning ownership to {instance.user}")
            CharacterOwnership.objects.update_or_create(character=char, defaults={'owner_hash': instance.character_owner_hash, 'user': instance.user})


@receiver(pre_delete, sender=CharacterOwnership)
def validate_main_character(sender, instance, *args, **kwargs):
    try:
        if instance.user.profile.main_character == instance.character:
            logger.info("Ownership of a main character {} has been revoked. Resetting {} main character.".format(
                instance.character, instance.user))
            # clear main character as user no longer owns them
            instance.user.profile.main_character = None
            instance.user.profile.save()
    except UserProfile.DoesNotExist:
        # a user is being deleted
        pass


@receiver(post_delete, sender=Token)
def validate_ownership(sender, instance, *args, **kwargs):
    if not Token.objects.filter(character_owner_hash=instance.character_owner_hash).filter(refresh_token__isnull=False).exists():
        logger.info(f"No remaining tokens to validate ownership of character {instance.character_name}. Revoking ownership.")
        CharacterOwnership.objects.filter(owner_hash=instance.character_owner_hash).delete()


@receiver(pre_save, sender=User)
def assign_state_on_active_change(sender, instance, *args, **kwargs):
    # set to guest state if inactive, assign proper state if reactivated
    if instance.pk:
        old_instance = User.objects.get(pk=instance.pk)
        if old_instance.is_active != instance.is_active:
            if instance.is_active:
                logger.debug(f"User {instance} has been activated. Assigning state.")
                instance.profile.assign_state()
            else:
                logger.debug(
                    f"User {instance} has been deactivated. Revoking state and assigning to guest state.")
                instance.profile.state = get_guest_state()
                instance.profile.save(update_fields=['state'])


@receiver(post_save, sender=EveCharacter)
def check_state_on_character_update(sender, instance, *args, **kwargs):
    # if this is a main character updating, check that user's state
    try:
        logger.debug(f"Character {instance} has been saved. Assessing owner's state for changes.")
        instance.userprofile.assign_state()
    except UserProfile.DoesNotExist:
        logger.debug(f"Character {instance} is not a main character. No state assessment required.")
        pass


@receiver(post_save, sender=CharacterOwnership)
def ownership_record_creation(sender, instance, created, *args, **kwargs):
    if created:
        records = OwnershipRecord.objects.filter(owner_hash=instance.owner_hash).filter(character=instance.character)
        if records.exists():
            if records[0].user == instance.user:  # most recent record is sorted first
                logger.debug(f"Already have ownership record of {instance.character} by user {instance.user}")
                return
        logger.info(f"Character {instance.character} has a new owner {instance.user}. Creating ownership record.")
        OwnershipRecord.objects.create(user=instance.user, character=instance.character, owner_hash=instance.owner_hash)



from django.contrib import admin
from .models import Ips4User


@admin.register(Ips4User)
class Ips4UserAdmin(admin.ModelAdmin):
        list_display = ('user', 'username', 'id')
        search_fields = ('user__username', 'username', 'id')



# Generated by Django 4.0.10 on 2023-10-07 07:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('authentication', '0021_alter_userprofile_language'),
    ]

    operations = [
        migrations.AddField(
            model_name='userprofile',
            name='theme',
            field=models.CharField(blank=True, help_text='Bootstrap 5 Themes from https://bootswatch.com/ or Community Apps', max_length=200, null=True, verbose_name='Theme'),
        ),
    ]



# Generated by Django 1.11.6 on 2017-12-23 04:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('authentication', '0015_user_profiles'),
        ('auth', '0008_alter_user_username_max_length'),
        ('eveonline', '0009_on_delete'),
    ]

    operations = [
        migrations.CreateModel(
            name='AutogroupsConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('corp_groups', models.BooleanField(default=False, help_text='Setting this to false will delete all the created groups.')),
                ('corp_group_prefix', models.CharField(blank=True, default='Corp ', max_length=50)),
                ('corp_name_source', models.CharField(choices=[('ticker', 'Ticker'), ('name', 'Full name')], default='name', max_length=20)),
                ('alliance_groups', models.BooleanField(default=False, help_text='Setting this to false will delete all the created groups.')),
                ('alliance_group_prefix', models.CharField(blank=True, default='Alliance ', max_length=50)),
                ('alliance_name_source', models.CharField(choices=[('ticker', 'Ticker'), ('name', 'Full name')], default='name', max_length=20)),
                ('replace_spaces', models.BooleanField(default=False)),
                ('replace_spaces_with', models.CharField(blank=True, default='', help_text='Any spaces in the group name will be replaced with this.', max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='ManagedAllianceGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('alliance', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveonline.EveAllianceInfo')),
                ('config', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eve_autogroups.AutogroupsConfig')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.Group')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ManagedCorpGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('config', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eve_autogroups.AutogroupsConfig')),
                ('corp', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveonline.EveCorporationInfo')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.Group')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='autogroupsconfig',
            name='alliance_managed_groups',
            field=models.ManyToManyField(help_text="A list of alliance groups created and maintained by this AutogroupConfig. You should not edit this list unless you know what you're doing.", related_name='alliance_managed_config', through='eve_autogroups.ManagedAllianceGroup', to='auth.Group'),
        ),
        migrations.AddField(
            model_name='autogroupsconfig',
            name='corp_managed_groups',
            field=models.ManyToManyField(help_text="A list of corporation groups created and maintained by this AutogroupConfig. You should not edit this list unless you know what you're doing.", related_name='corp_managed_config', through='eve_autogroups.ManagedCorpGroup', to='auth.Group'),
        ),
        migrations.AddField(
            model_name='autogroupsconfig',
            name='states',
            field=models.ManyToManyField(related_name='autogroups', to='authentication.State'),
        ),
    ]



from django.apps import apps
from allianceauth.authentication.models import User
from esi.models import Token


def install_stat_users() -> int:
    """Count and Return the number of User accounts

    Returns
    -------
    int
        The Number of User objects"""
    users = User.objects.count()
    return users


def install_stat_tokens() -> int:
    """Count and Return the number of ESI Tokens Stored

    Returns
    -------
    int
        The Number of Token Objects"""
    tokens = Token.objects.count()
    return tokens


def install_stat_addons() -> int:
    """Count and Return the number of Django Applications Installed

    Returns
    -------
    int
        The Number of Installed Apps"""
    addons = len(list(apps.get_app_configs()))
    return addons



from django.urls import path
from django.views.generic.base import TemplateView

from . import views

app_name = 'authentication'

urlpatterns = [
    path('', views.index, name='index'),
    path(
        'account/login/',
        TemplateView.as_view(template_name='public/login.html'),
        name='login'
    ),
    path(
        'account/characters/main/',
        views.main_character_change,
        name='change_main_character'
    ),
    path(
        'account/characters/add/',
        views.add_character,
        name='add_character'
    ),
    path(
        'account/tokens/manage/',
        views.token_management,
        name='token_management'
    ),
    path(
        'account/tokens/delete/<int:token_id>',
        views.token_delete,
        name='token_delete'
    ),
    path(
        'account/tokens/refresh/<int:token_id>',
        views.token_refresh,
        name='token_refresh'
    ),
    path('dashboard/', views.dashboard, name='dashboard'),
    path('task-counts/', views.task_counts, name='task_counts'),
]



from django.urls import include
from allianceauth.hooks import get_hooks
from django.urls import path

from . import views

urlpatterns = [
    # Services
    path('services/', include(([
        path('', views.services_view, name='services'),
        # Tools
        path('tool/fleet_formatter_tool/', views.fleet_formatter_view, name='fleet_format_tool'),
    ], 'services'), namespace='services')),
]

# Append hooked service urls
services = get_hooks('services_hook')
for svc in services:
    urlpatterns += svc().urlpatterns



# Generated by Django 4.0.6 on 2022-08-30 05:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('analytics', '0006_more_ignore_paths'),
    ]

    operations = [
        migrations.AddField(
            model_name='analyticstokens',
            name='secret',
            field=models.CharField(blank=True, max_length=254),
        ),
    ]



from django.urls import include, path

from . import views

app_name = 'ips4'

module_urls = [
    # IPS4 Service Control
    path('activate/', views.activate_ips4, name='activate'),
    path('deactivate/', views.deactivate_ips4, name='deactivate'),
    path('reset_password/', views.reset_ips4_password, name='reset_password'),
    path('set_password/', views.set_ips4_password, name='set_password'),
]

urlpatterns = [
    path('ips4/', include((module_urls, app_name), namespace=app_name))
]



# Generated by Django 3.1.1 on 2020-09-18 14:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hrapplications', '0006_remove_legacy_models'),
    ]

    operations = [
        migrations.AlterField(
            model_name='application',
            name='approved',
            field=models.BooleanField(blank=True, default=None, null=True),
        ),
    ]



import logging
from typing import Union

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from esi.models import Token

from allianceauth.notifications import notify

from . import providers
from .evelinks import eveimageserver
from .managers import (
    EveAllianceManager,
    EveAllianceProviderManager,
    EveCharacterManager,
    EveCharacterProviderManager,
    EveCorporationManager,
    EveCorporationProviderManager,
)

logger = logging.getLogger(__name__)

_DEFAULT_IMAGE_SIZE = 32
DOOMHEIM_CORPORATION_ID = 1000001


class EveFactionInfo(models.Model):
    """A faction in Eve Online."""

    faction_id = models.PositiveIntegerField(unique=True, db_index=True)
    faction_name = models.CharField(max_length=254, unique=True)

    provider = providers.provider

    def __str__(self):
        return self.faction_name

    @staticmethod
    def generic_logo_url(
        faction_id: int, size: int = _DEFAULT_IMAGE_SIZE
    ) -> str:
        """image URL for the given faction ID"""
        return eveimageserver.corporation_logo_url(faction_id, size)

    def logo_url(self, size: int = _DEFAULT_IMAGE_SIZE) -> str:
        """image URL of this faction"""
        return self.generic_logo_url(self.faction_id, size)

    @property
    def logo_url_32(self) -> str:
        """image URL for this faction"""
        return self.logo_url(32)

    @property
    def logo_url_64(self) -> str:
        """image URL for this faction"""
        return self.logo_url(64)

    @property
    def logo_url_128(self) -> str:
        """image URL for this faction"""
        return self.logo_url(128)

    @property
    def logo_url_256(self) -> str:
        """image URL for this faction"""
        return self.logo_url(256)


class EveAllianceInfo(models.Model):
    """An alliance in Eve Online."""

    alliance_id = models.PositiveIntegerField(unique=True)
    alliance_name = models.CharField(max_length=254, db_index=True)
    alliance_ticker = models.CharField(max_length=254)
    executor_corp_id = models.PositiveIntegerField()

    objects = EveAllianceManager()
    provider = EveAllianceProviderManager()

    class Meta:
        indexes = [models.Index(fields=['executor_corp_id',])]

    def populate_alliance(self):
        alliance = self.provider.get_alliance(self.alliance_id)
        for corp_id in alliance.corp_ids:
            if not EveCorporationInfo.objects.filter(corporation_id=corp_id).exists():
                EveCorporationInfo.objects.create_corporation(corp_id)
        EveCorporationInfo.objects.filter(corporation_id__in=alliance.corp_ids).update(
            alliance=self
        )
        EveCorporationInfo.objects.filter(alliance=self).exclude(
            corporation_id__in=alliance.corp_ids
        ).update(alliance=None)

    def update_alliance(self, alliance: providers.Alliance = None):
        if alliance is None:
            alliance = self.provider.get_alliance(self.alliance_id)
        self.executor_corp_id = alliance.executor_corp_id
        self.save()
        return self

    def __str__(self):
        return self.alliance_name

    @staticmethod
    def generic_logo_url(
        alliance_id: int, size: int = _DEFAULT_IMAGE_SIZE
    ) -> str:
        """image URL for the given alliance ID"""
        return eveimageserver.alliance_logo_url(alliance_id, size)

    def logo_url(self, size: int = _DEFAULT_IMAGE_SIZE) -> str:
        """image URL of this alliance"""
        return self.generic_logo_url(self.alliance_id, size)

    @property
    def logo_url_32(self) -> str:
        """image URL for this alliance"""
        return self.logo_url(32)

    @property
    def logo_url_64(self) -> str:
        """image URL for this alliance"""
        return self.logo_url(64)

    @property
    def logo_url_128(self) -> str:
        """image URL for this alliance"""
        return self.logo_url(128)

    @property
    def logo_url_256(self) -> str:
        """image URL for this alliance"""
        return self.logo_url(256)


class EveCorporationInfo(models.Model):
    """A corporation in Eve Online."""

    corporation_id = models.PositiveIntegerField(unique=True)
    corporation_name = models.CharField(max_length=254, db_index=True)
    corporation_ticker = models.CharField(max_length=254)
    member_count = models.IntegerField()
    ceo_id = models.PositiveIntegerField(blank=True, null=True, default=None)
    alliance = models.ForeignKey(
        EveAllianceInfo, blank=True, null=True, on_delete=models.SET_NULL
    )

    objects = EveCorporationManager()
    provider = EveCorporationProviderManager()

    class Meta:
        indexes = [models.Index(fields=['ceo_id',]),]

    def update_corporation(self, corp: providers.Corporation = None):
        if corp is None:
            corp = self.provider.get_corporation(self.corporation_id)
        self.member_count = corp.members
        self.ceo_id = corp.ceo_id
        try:
            self.alliance = EveAllianceInfo.objects.get(alliance_id=corp.alliance_id)
        except EveAllianceInfo.DoesNotExist:
            self.alliance = None
        self.save()
        return self

    def __str__(self):
        return self.corporation_name

    @staticmethod
    def generic_logo_url(
        corporation_id: int, size: int = _DEFAULT_IMAGE_SIZE
    ) -> str:
        """image URL for the given corporation ID"""
        return eveimageserver.corporation_logo_url(corporation_id, size)

    def logo_url(self, size: int = _DEFAULT_IMAGE_SIZE) -> str:
        """image URL for this corporation"""
        return self.generic_logo_url(self.corporation_id, size)

    @property
    def logo_url_32(self) -> str:
        """image URL for this corporation"""
        return self.logo_url(32)

    @property
    def logo_url_64(self) -> str:
        """image URL for this corporation"""
        return self.logo_url(64)

    @property
    def logo_url_128(self) -> str:
        """image URL for this corporation"""
        return self.logo_url(128)

    @property
    def logo_url_256(self) -> str:
        """image URL for this corporation"""
        return self.logo_url(256)


class EveCharacter(models.Model):
    """A character in Eve Online."""

    character_id = models.PositiveIntegerField(unique=True)
    character_name = models.CharField(max_length=254, db_index=True)
    corporation_id = models.PositiveIntegerField()
    corporation_name = models.CharField(max_length=254)
    corporation_ticker = models.CharField(max_length=5)
    alliance_id = models.PositiveIntegerField(blank=True, null=True, default=None)
    alliance_name = models.CharField(max_length=254, blank=True, null=True, default='')
    alliance_ticker = models.CharField(max_length=5, blank=True, null=True, default='')
    faction_id = models.PositiveIntegerField(blank=True, null=True, default=None)
    faction_name = models.CharField(max_length=254, blank=True, null=True, default='')

    objects = EveCharacterManager()
    provider = EveCharacterProviderManager()

    class Meta:
        indexes = [
            models.Index(fields=['corporation_id',]),
            models.Index(fields=['alliance_id',]),
            models.Index(fields=['corporation_name',]),
            models.Index(fields=['alliance_name',]),
            models.Index(fields=['faction_id',]),
        ]

    def __str__(self):
        return self.character_name

    @property
    def is_biomassed(self) -> bool:
        """Whether this character is dead or not."""
        return self.corporation_id == DOOMHEIM_CORPORATION_ID

    @property
    def alliance(self) -> Union[EveAllianceInfo, None]:
        """
        Pseudo foreign key from alliance_id to EveAllianceInfo
        :raises: EveAllianceInfo.DoesNotExist
        :return: EveAllianceInfo or None
        """
        if self.alliance_id is None:
            return None
        return EveAllianceInfo.objects.get(alliance_id=self.alliance_id)

    @property
    def corporation(self) -> EveCorporationInfo:
        """
        Pseudo foreign key from corporation_id to EveCorporationInfo
        :raises: EveCorporationInfo.DoesNotExist
        :return: EveCorporationInfo
        """
        return EveCorporationInfo.objects.get(corporation_id=self.corporation_id)

    @property
    def faction(self) -> Union[EveFactionInfo, None]:
        """
        Pseudo foreign key from faction_id to EveFactionInfo
        :raises: EveFactionInfo.DoesNotExist
        :return: EveFactionInfo
        """
        if self.faction_id is None:
            return None
        return EveFactionInfo.objects.get(faction_id=self.faction_id)

    def update_character(self, character: providers.Character = None):
        if character is None:
            character = self.provider.get_character(self.character_id)
        self.character_name = character.name
        self.corporation_id = character.corp.id
        self.corporation_name = character.corp.name
        self.corporation_ticker = character.corp.ticker
        self.alliance_id = character.alliance.id
        self.alliance_name = character.alliance.name
        self.alliance_ticker = getattr(character.alliance, 'ticker', None)
        self.faction_id = character.faction.id
        self.faction_name = character.faction.name
        self.save()
        if self.is_biomassed:
            self._remove_tokens_of_biomassed_character()
        return self

    def _remove_tokens_of_biomassed_character(self) -> None:
        """Remove tokens of this biomassed character."""
        try:
            user = self.character_ownership.user
        except ObjectDoesNotExist:
            return
        tokens_to_delete = Token.objects.filter(character_id=self.character_id)
        tokens_count = tokens_to_delete.count()
        if not tokens_count:
            return
        tokens_to_delete.delete()
        logger.info(
            "%d tokens from user %s for biomassed character %s [id:%s] deleted.",
            tokens_count,
            user,
            self,
            self.character_id,
        )
        notify(
            user=user,
            title=f"Character {self} biomassed",
            message=(
                f"Your former character {self} has been biomassed "
                "and has been removed from the list of your alts."
            )
        )

    @staticmethod
    def generic_portrait_url(
        character_id: int, size: int = _DEFAULT_IMAGE_SIZE
    ) -> str:
        """image URL for the given character ID"""
        return eveimageserver.character_portrait_url(character_id, size)

    def portrait_url(self, size=_DEFAULT_IMAGE_SIZE) -> str:
        """image URL for this character"""
        return self.generic_portrait_url(self.character_id, size)

    @property
    def portrait_url_32(self) -> str:
        """image URL for this character"""
        return self.portrait_url(32)

    @property
    def portrait_url_64(self) -> str:
        """image URL for this character"""
        return self.portrait_url(64)

    @property
    def portrait_url_128(self) -> str:
        """image URL for this character"""
        return self.portrait_url(128)

    @property
    def portrait_url_256(self) -> str:
        """image URL for this character"""
        return self.portrait_url(256)

    def corporation_logo_url(self, size=_DEFAULT_IMAGE_SIZE) -> str:
        """image URL for corporation of this character"""
        return EveCorporationInfo.generic_logo_url(self.corporation_id, size)

    @property
    def corporation_logo_url_32(self) -> str:
        """image URL for corporation of this character"""
        return self.corporation_logo_url(32)

    @property
    def corporation_logo_url_64(self) -> str:
        """image URL for corporation of this character"""
        return self.corporation_logo_url(64)

    @property
    def corporation_logo_url_128(self) -> str:
        """image URL for corporation of this character"""
        return self.corporation_logo_url(128)

    @property
    def corporation_logo_url_256(self) -> str:
        """image URL for corporation of this character"""
        return self.corporation_logo_url(256)

    def alliance_logo_url(self, size=_DEFAULT_IMAGE_SIZE) -> str:
        """image URL for alliance of this character or empty string"""
        if self.alliance_id:
            return EveAllianceInfo.generic_logo_url(self.alliance_id, size)
        else:
            return ''

    @property
    def alliance_logo_url_32(self) -> str:
        """image URL for alliance of this character or empty string"""
        return self.alliance_logo_url(32)

    @property
    def alliance_logo_url_64(self) -> str:
        """image URL for alliance of this character or empty string"""
        return self.alliance_logo_url(64)

    @property
    def alliance_logo_url_128(self) -> str:
        """image URL for alliance of this character or empty string"""
        return self.alliance_logo_url(128)

    @property
    def alliance_logo_url_256(self) -> str:
        """image URL for alliance of this character or empty string"""
        return self.alliance_logo_url(256)

    def faction_logo_url(self, size=_DEFAULT_IMAGE_SIZE) -> str:
        """image URL for alliance of this character or empty string"""
        if self.faction_id:
            return EveFactionInfo.generic_logo_url(self.faction_id, size)
        else:
            return ''

    @property
    def faction_logo_url_32(self) -> str:
        """image URL for alliance of this character or empty string"""
        return self.faction_logo_url(32)

    @property
    def faction_logo_url_64(self) -> str:
        """image URL for alliance of this character or empty string"""
        return self.faction_logo_url(64)

    @property
    def faction_logo_url_128(self) -> str:
        """image URL for alliance of this character or empty string"""
        return self.faction_logo_url(128)

    @property
    def faction_logo_url_256(self) -> str:
        """image URL for alliance of this character or empty string"""
        return self.faction_logo_url(256)



from itertools import count

from django.utils.timezone import now

from ..client import DiscordApiStatusCode
from ..models import Guild, GuildMember, Role, User

TEST_GUILD_ID = 123456789012345678
TEST_GUILD_NAME = "Test Guild"
TEST_USER_ID = 198765432012345678
TEST_USER_NAME = "Peter Parker"
TEST_USER_DISCRIMINATOR = "1234"
TEST_BOT_TOKEN = "abcdefhijlkmnopqastzvwxyz1234567890ABCDEFGHOJKLMNOPQRSTUVWXY"
TEST_ROLE_ID = 654321012345678912


def create_discord_role_object(id: int, name: str, managed: bool = False) -> dict:
    return {"id": str(int(id)), "name": str(name), "managed": bool(managed)}


def create_matched_role(role, created=False) -> tuple:
    return role, created


def create_discord_user_object(**kwargs):
    params = {
        "id": TEST_USER_ID,
        "username": TEST_USER_NAME,
        "discriminator": TEST_USER_DISCRIMINATOR,
    }
    params.update(kwargs)
    params["id"] = str(int(params["id"]))
    return params


def create_discord_guild_member_object(user=None, **kwargs):
    user_params = {}
    if user:
        user_params["user"] = user
    params = {
        "user": create_discord_user_object(**user_params),
        "roles": [],
        "joined_at": now().isoformat(),
        "deaf": False,
        "mute": False,
    }
    params.update(kwargs)
    params["roles"] = [str(int(obj)) for obj in params["roles"]]
    return params


def create_discord_error_response(code: int) -> dict:
    return {"code": int(code)}


def create_discord_error_response_unknown_member() -> dict:
    return create_discord_error_response(DiscordApiStatusCode.UNKNOWN_MEMBER.value)


def create_discord_guild_object(**kwargs):
    params = {"id": TEST_GUILD_ID, "name": TEST_GUILD_NAME, "roles": []}
    params.update(kwargs)
    params["id"] = str(int(params["id"]))
    return params


def create_user(**kwargs):
    params = {
        "id": TEST_USER_ID,
        "username": TEST_USER_NAME,
        "discriminator": TEST_USER_DISCRIMINATOR,
    }
    params.update(kwargs)
    return User(**params)


def create_guild(**kwargs):
    params = {"id": TEST_GUILD_ID, "name": TEST_GUILD_NAME, "roles": []}
    params.update(kwargs)
    return Guild(**params)


def create_guild_member(**kwargs):
    params = {"user": create_user(), "roles": []}
    params.update(kwargs)
    return GuildMember(**params)


def create_role(**kwargs) -> dict:
    params = {"managed": False}
    params.update(kwargs)
    if "id" not in params:
        params["id"] = next_number("role")
    if "name" not in params:
        params["name"] = f"Test Role #{params['id']}"
    return Role(**params)


def next_number(key: str = None) -> int:
    """Calculate the next number in a persistent sequence."""
    if key is None:
        key = "_general"
    try:
        return next_number._counter[key].__next__()
    except AttributeError:
        next_number._counter = dict()
    except KeyError:
        pass
    next_number._counter[key] = count(start=1)
    return next_number._counter[key].__next__()



"""Implementation of Discord objects used by this client.

Note that only those objects and properties are implemented, which are needed by AA.

Names and types are mirrored from the API whenever possible.
Discord's snowflake type (used by Discord IDs) is implemented as int.
"""

from dataclasses import asdict, dataclass
from typing import FrozenSet


@dataclass(frozen=True)
class User:
    """A user on Discord."""

    id: int
    username: str
    discriminator: str

    def __post_init__(self):
        object.__setattr__(self, "id", int(self.id))
        object.__setattr__(self, "username", str(self.username))
        object.__setattr__(self, "discriminator", str(self.discriminator))

    @classmethod
    def from_dict(cls, data: dict) -> "User":
        """Create object from dictionary as received from the API."""
        return cls(
            id=int(data["id"]),
            username=data["username"],
            discriminator=data["discriminator"],
        )


@dataclass(frozen=True)
class Role:
    """A role on Discord."""

    _ROLE_NAME_MAX_CHARS = 100

    id: int
    name: str
    managed: bool = False

    def __post_init__(self):
        object.__setattr__(self, "id", int(self.id))
        object.__setattr__(self, "name", self.sanitize_name(self.name))
        object.__setattr__(self, "managed", bool(self.managed))

    def asdict(self) -> dict:
        """Convert object into a dictionary representation."""
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict) -> "Role":
        """Create object from dictionary as received from the API."""
        return cls(id=int(data["id"]), name=data["name"], managed=data["managed"])

    @classmethod
    def sanitize_name(cls, role_name: str) -> str:
        """Shorten too long names if necessary."""
        return str(role_name)[: cls._ROLE_NAME_MAX_CHARS]


@dataclass(frozen=True)
class Guild:
    """A guild on Discord."""

    id: int
    name: str
    roles: FrozenSet[Role]

    def __post_init__(self):
        object.__setattr__(self, "id", int(self.id))
        object.__setattr__(self, "name", str(self.name))
        for role in self.roles:
            if not isinstance(role, Role):
                raise TypeError("roles can only contain Role objects.")
        object.__setattr__(self, "roles", frozenset(self.roles))

    @classmethod
    def from_dict(cls, data: dict) -> "Guild":
        """Create object from dictionary as received from the API."""
        return cls(
            id=int(data["id"]),
            name=data["name"],
            roles=frozenset(Role.from_dict(obj) for obj in data["roles"]),
        )


@dataclass(frozen=True)
class GuildMember:
    """A member of a guild on Discord."""

    _NICK_MAX_CHARS = 32

    roles: FrozenSet[int]
    nick: str = None
    user: User = None

    def __post_init__(self):
        if self.nick:
            object.__setattr__(self, "nick", self.sanitize_nick(self.nick))
        if self.user and not isinstance(self.user, User):
            raise TypeError("user must be of type User")
        for role in self.roles:
            if not isinstance(role, int):
                raise TypeError("roles can only contain ints")
        object.__setattr__(self, "roles", frozenset(self.roles))

    @classmethod
    def from_dict(cls, data: dict) -> "GuildMember":
        """Create object from dictionary as received from the API."""
        params = {"roles": {int(obj) for obj in data["roles"]}}
        if data.get("user"):
            params["user"] = User.from_dict(data["user"])
        if data.get("nick"):
            params["nick"] = data["nick"]
        return cls(**params)

    @classmethod
    def sanitize_nick(cls, nick: str) -> str:
        """Sanitize a nick, i.e. shorten too long strings if necessary."""
        return str(nick)[: cls._NICK_MAX_CHARS]



# Generated by Django 1.10.5 on 2017-03-22 23:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('eveonline', '0007_unique_id_name'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='eveapikeypair',
            name='user',
        ),
        migrations.RemoveField(
            model_name='eveallianceinfo',
            name='is_blue',
        ),
        migrations.RemoveField(
            model_name='evecharacter',
            name='api_id',
        ),
        migrations.RemoveField(
            model_name='evecharacter',
            name='user',
        ),
        migrations.RemoveField(
            model_name='evecorporationinfo',
            name='is_blue',
        ),
        migrations.DeleteModel(
            name='EveApiKeyPair',
        ),
    ]



# this module generates profile URLs for dotlan

from urllib.parse import urljoin, quote

from . import (
    _ESI_CATEGORY_ALLIANCE,
    _ESI_CATEGORY_CORPORATION,
    _ESI_CATEGORY_REGION,
    _ESI_CATEGORY_SOLARSYSTEM
)


_BASE_URL = 'http://evemaps.dotlan.net'


def _build_url(category: str, name: str) -> str:
    """return url to profile page for an eve entity"""

    if category == _ESI_CATEGORY_ALLIANCE:
        partial = 'alliance'

    elif category == _ESI_CATEGORY_CORPORATION:
        partial = 'corp'

    elif category == _ESI_CATEGORY_REGION:
        partial = 'map'

    elif category == _ESI_CATEGORY_SOLARSYSTEM:
        partial = 'system'

    else:
        raise NotImplementedError(
            "Not implemented yet for category:" + category
        )

    url = urljoin(
        _BASE_URL,
        '{}/{}'.format(partial, quote(str(name).replace(" ", "_")))

    )
    return url


def alliance_url(name: str) -> str:
    """url for page about given alliance on dotlan"""
    return _build_url(_ESI_CATEGORY_ALLIANCE, name)


def corporation_url(name: str) -> str:
    """url for page about given corporation on dotlan"""
    return _build_url(_ESI_CATEGORY_CORPORATION, name)


def region_url(name: str) -> str:
    """url for page about given region on dotlan"""
    return _build_url(_ESI_CATEGORY_REGION, name)


def solar_system_url(name: str) -> str:
    """url for page about given solar system on dotlan"""
    return _build_url(_ESI_CATEGORY_SOLARSYSTEM, name)



from celery import shared_task

from django.contrib.auth.models import Group


@shared_task
def remove_users_not_matching_states_from_group(group_pk: int) -> None:
    """Remove users not matching defined states from related group."""
    group = Group.objects.get(pk=group_pk)
    group.authgroup.remove_users_not_matching_states()



import logging
import random
import string
import re
from django.db import connections
from passlib.hash import bcrypt
from django.conf import settings

logger = logging.getLogger(__name__)


TABLE_PREFIX = getattr(settings, 'IPS4_TABLE_PREFIX', '')


class Ips4Manager:
    SQL_ADD_USER = r"INSERT INTO %score_members (name, email, members_pass_hash, members_pass_salt, " \
                    r"member_group_id) VALUES (%%s, %%s, %%s, %%s, %%s)" % TABLE_PREFIX
    SQL_GET_ID = r"SELECT member_id FROM %score_members WHERE name = %%s" % TABLE_PREFIX
    SQL_UPDATE_PASSWORD = r"UPDATE %score_members SET members_pass_hash = %%s, members_pass_salt = %%s WHERE name = %%s" % TABLE_PREFIX
    SQL_DEL_USER = r"DELETE FROM %score_members WHERE member_id = %%s" % TABLE_PREFIX

    MEMBER_GROUP_ID = 3

    @classmethod
    def add_user(cls, username, email):
        logger.debug("Adding new IPS4 user %s" % username)
        plain_password = cls.__generate_random_pass()
        hash = cls._gen_pwhash(plain_password)
        salt = cls._get_salt(hash)
        group = cls.MEMBER_GROUP_ID
        cursor = connections['ips4'].cursor()
        cursor.execute(cls.SQL_ADD_USER, [username, email, hash, salt, group])
        member_id = cls.get_user_id(username)
        return username, plain_password, member_id

    @staticmethod
    def get_user_id(username):
        cursor = connections['ips4'].cursor()
        cursor.execute(Ips4Manager.SQL_GET_ID, [username])
        row = cursor.fetchone()
        if row is not None:
            logger.debug(f"Got user id {row[0]} for username {username}")
            return row[0]
        else:
            logger.error("username %s not found. Unable to determine id." % username)
            return None

    @staticmethod
    def __generate_random_pass():
        return ''.join([random.choice(string.ascii_letters + string.digits) for n in range(16)])

    @staticmethod
    def _gen_pwhash(password):
        return bcrypt.using(ident='2y').encrypt(password.encode('utf-8'), rounds=13)

    @staticmethod
    def _get_salt(pw_hash):
        search = re.compile(r"^\$2[a-z]?\$([0-9]+)\$(.{22})(.{31})$")
        match = re.match(search, pw_hash)
        return match.group(2)

    @staticmethod
    def delete_user(id):
        logger.debug("Deleting IPS4 user id %s" % id)
        try:
            cursor = connections['ips4'].cursor()
            cursor.execute(Ips4Manager.SQL_DEL_USER, [id])
            logger.info("Deleted IPS4 user %s" % id)
            return True
        except:
            logger.exception("Failed to delete IPS4 user id %s" % id)
            return False

    @classmethod
    def update_user_password(cls, username):
        logger.debug("Updating IPS4 user id %s password" % id)
        if cls.check_user(username):
            plain_password = Ips4Manager.__generate_random_pass()
            hash = cls._gen_pwhash(plain_password)
            salt = cls._get_salt(hash)
            cursor = connections['ips4'].cursor()
            cursor.execute(cls.SQL_UPDATE_PASSWORD, [hash, salt, username])
            return plain_password
        else:
            logger.error("Unable to update ips4 user %s password" % username)
            return ""

    @staticmethod
    def check_user(username):
        logger.debug("Checking IPS4 username %s" % username)
        cursor = connections['ips4'].cursor()
        cursor.execute(Ips4Manager.SQL_GET_ID, [username])
        row = cursor.fetchone()
        if row:
            logger.debug("Found user %s on IPS4" % username)
            return True
        logger.debug("User %s not found on IPS4" % username)
        return False

    @classmethod
    def update_custom_password(cls, username, plain_password):
        logger.debug("Updating IPS4 user id %s password" % id)
        if cls.check_user(username):
            hash = cls._gen_pwhash(plain_password)
            salt = cls._get_salt(hash)
            cursor = connections['ips4'].cursor()
            cursor.execute(cls.SQL_UPDATE_PASSWORD, [hash, salt, username])
            return plain_password
        else:
            logger.error("Unable to update ips4 user %s password" % username)
            return ""



from django.apps import AppConfig


class FlatlyThemeConfig(AppConfig):
    name = "allianceauth.theme.flatly"
    label = "flatly"
    version = "5.3.0"
    verbose_name = f"Bootswatch Flatly v{version}"

    def ready(self):
        pass



from bravado.exception import HTTPNotFound


class BravadoResponseStub:
    """Stub for IncomingResponse in bravado, e.g. for HTTPError exceptions"""

    def __init__(
        self, status_code, reason="", text="", headers=None, raw_bytes=None
    ) -> None:
        self.reason = reason
        self.status_code = status_code
        self.text = text
        self.headers = headers if headers else dict()
        self.raw_bytes = raw_bytes

    def __str__(self):
        return f"{self.status_code} {self.reason}"


class BravadoOperationStub:
    """Stub to simulate the operation object return from bravado via django-esi"""

    class RequestConfig:
        def __init__(self, also_return_response):
            self.also_return_response = also_return_response

    class ResponseStub:
        def __init__(self, headers):
            self.headers = headers

    def __init__(self, data, headers: dict = None, also_return_response: bool = False):
        self._data = data
        self._headers = headers if headers else {"x-pages": 1}
        self.request_config = BravadoOperationStub.RequestConfig(also_return_response)

    def result(self, **kwargs):
        if self.request_config.also_return_response:
            return [self._data, self.ResponseStub(self._headers)]
        else:
            return self._data

    def results(self, **kwargs):
        return self.result(**kwargs)


class EsiClientStub:
    """Stub for an ESI client."""
    class Alliance:
        @staticmethod
        def get_alliances_alliance_id(alliance_id):
            data = {
                3001: {
                    "name": "Wayne Enterprises",
                    "ticker": "WYE",
                    "executor_corporation_id": 2001
                }
            }
            try:
                return BravadoOperationStub(data[int(alliance_id)])
            except KeyError:
                response = BravadoResponseStub(
                    404, f"Alliance with ID {alliance_id} not found"
                )
                raise HTTPNotFound(response)

        @staticmethod
        def get_alliances_alliance_id_corporations(alliance_id):
            data = [2001, 2002, 2003]
            return BravadoOperationStub(data)

    class Character:
        @staticmethod
        def get_characters_character_id(character_id):
            data = {
                1001: {
                    "corporation_id": 2001,
                    "name": "Bruce Wayne",
                },
                1002: {
                    "corporation_id": 2001,
                    "name": "Peter Parker",
                },
                1011: {
                    "corporation_id": 2011,
                    "name": "Lex Luthor",
                }
            }
            try:
                return BravadoOperationStub(data[int(character_id)])
            except KeyError:
                response = BravadoResponseStub(
                    404, f"Character with ID {character_id} not found"
                )
                raise HTTPNotFound(response)

        @staticmethod
        def post_characters_affiliation(characters: list):
            data = [
                {'character_id': 1001, 'corporation_id': 2001, 'alliance_id': 3001},
                {'character_id': 1002, 'corporation_id': 2001, 'alliance_id': 3001},
                {'character_id': 1011, 'corporation_id': 2011},
                {'character_id': 1666, 'corporation_id': 1000001},
            ]
            return BravadoOperationStub(
                [x for x in data if x['character_id'] in characters]
            )

    class Corporation:
        @staticmethod
        def get_corporations_corporation_id(corporation_id):
            data = {
                2001: {
                    "ceo_id": 1091,
                    "member_count": 10,
                    "name": "Wayne Technologies",
                    "ticker": "WTE",
                    "alliance_id": 3001
                },
                2002: {
                    "ceo_id": 1092,
                    "member_count": 10,
                    "name": "Wayne Food",
                    "ticker": "WFO",
                    "alliance_id": 3001
                },
                2003: {
                    "ceo_id": 1093,
                    "member_count": 10,
                    "name": "Wayne Energy",
                    "ticker": "WEG",
                    "alliance_id": 3001
                },
                2011: {
                    "ceo_id": 1,
                    "member_count": 3,
                    "name": "LexCorp",
                    "ticker": "LC",
                },
                1000001: {
                    "ceo_id": 3000001,
                    "creator_id": 1,
                    "description": "The internal corporation used for characters in graveyard.",
                    "member_count": 6329026,
                    "name": "Doomheim",
                    "ticker": "666",
                }
            }
            try:
                return BravadoOperationStub(data[int(corporation_id)])
            except KeyError:
                response = BravadoResponseStub(
                    404, f"Corporation with ID {corporation_id} not found"
                )
                raise HTTPNotFound(response)

    class Universe:
        @staticmethod
        def post_universe_names(ids: list):
            data = [
                {"category": "character", "id": 1001, "name": "Bruce Wayne"},
                {"category": "character", "id": 1002, "name": "Peter Parker"},
                {"category": "character", "id": 1011, "name": "Lex Luthor"},
                {"category": "character", "id": 1666, "name": "Hal Jordan"},
                {"category": "corporation", "id": 2001, "name": "Wayne Technologies"},
                {"category": "corporation","id": 2002, "name": "Wayne Food"},
                {"category": "corporation","id": 1000001, "name": "Doomheim"},
            ]
            return BravadoOperationStub([x for x in data if x['id'] in ids])



import logging
from allianceauth.hooks import get_hooks

from django.db import models
from django.utils.translation import gettext_lazy as _
from django.template.loader import render_to_string

logger = logging.getLogger(__name__)


class MenuItem(models.Model):
    # Auto Generated model from an auth_hook
    hook_function = models.CharField(
        max_length=500, default=None, null=True, blank=True)

    # User Made Model
    icon_classes = models.CharField(
        max_length=150, default=None, null=True, blank=True, help_text="Font Awesome classes to show as icon on menu")
    text = models.CharField(
        max_length=150, default=None, null=True, blank=True, help_text="Text to show on menu")
    url = models.CharField(max_length=2048, default=None,
                            null=True, blank=True)

    # Put it under a header?
    parent = models.ForeignKey(
        'self', on_delete=models.SET_NULL, null=True, blank=True, help_text="Parent Header. (Optional)")

    # Put it where? lowest first
    rank = models.IntegerField(default=1000, help_text="Order of the menu. Lowest First.")

    # Hide it fully? Hiding a parent will hide all it's children
    hide = models.BooleanField(default=False, help_text="Hide this menu item. If this item is a header all items under it will be hidden too.")

    class Meta:
        indexes = [
            models.Index(fields=['rank', ]),
        ]

    def __str__(self) -> str:
        return self.text

    @property
    def classes(self):  # Helper function to make this model closer to the hook functions
        return self.icon_classes

    @staticmethod
    def hook_to_name(mh):
        return f"{mh.__class__.__module__}.{mh.__class__.__name__}"

    @staticmethod
    def sync_hook_models():
        # TODO define aa way for hooks to predefine a "parent" to create a sub menu from modules
        menu_hooks = get_hooks('menu_item_hook')
        hook_functions = []
        for hook in menu_hooks:
            mh = hook()
            cls = MenuItem.hook_to_name(mh)
            try:
                # if it exists update the text only
                # Users can adjust ranks so lets not change it if they have.
                mi = MenuItem.objects.get(hook_function=cls)
                mi.text = getattr(mh, "text", mh.__class__.__name__)
                mi.save()
            except MenuItem.DoesNotExist:
                # This is a new hook, Make the database model.
                MenuItem.objects.create(
                    hook_function=cls,
                    rank=getattr(mh, "order", 500),
                    text=getattr(mh, "text", mh.__class__.__name__)
                )
            hook_functions.append(cls)

        # Get rid of any legacy hooks from modules removed
        MenuItem.objects.filter(hook_function__isnull=False).exclude(
            hook_function__in=hook_functions).delete()

    @classmethod
    def filter_items(cls, menu_item: dict):
        """
        filter any items with no valid children from a menu
        """
        count_items = len(menu_item['items'])
        if count_items:  # if we have children confirm we can see them
            for i in menu_item['items']:
                if len(i['render']) == 0:
                    count_items -= 1
            if count_items == 0:  # no children left dont render header
                return False
            return True
        else:
            return True

    @classmethod
    def render_menu(cls, request):
        """
        Return the sorted side menu items with any items the user can't see removed.
        """
        # Override all the items to the bs5 theme
        template = "menu/menu-item-bs5.html"
        # TODO discuss permissions for user defined links

        # Turn all the hooks into functions
        menu_hooks = get_hooks('menu_item_hook')
        items = {}
        for fn in menu_hooks:
            f = fn()
            items[cls.hook_to_name(f)] = f

        menu_items = MenuItem.objects.all().order_by("rank")

        menu = {}
        for mi in menu_items:
            if mi.hide:
                # hidden item, skip it completely
                continue
            try:
                _cnt = 0
                _render = None
                if mi.hook_function:
                    # This is a module hook, so we need to render it as the developer intended
                    # TODO add a new attribute for apps that want to override it in the new theme
                    items[mi.hook_function].template = template
                    _render = items[mi.hook_function].render(request)
                    _cnt = items[mi.hook_function].count
                else:
                    # This is a user defined menu item so we render it with defaults.
                    _render = render_to_string(template,
                                                {'item': mi},
                                                request=request)

                parent = mi.id
                if mi.parent_id:  # Set it if present
                    parent = mi.parent_id

                if parent not in menu:  # this will cause the menu headers to be out of order
                    menu[parent] = {"items": [],
                                    "count": 0,
                                    "render": None,
                                    "text": "None",
                                    "rank": 9999,
                                    }
                _mi = {
                    "count": _cnt,
                    "render": _render,
                    "text": mi.text,
                    "rank": mi.rank,
                    "classes": (mi.icon_classes if mi.icon_classes != "" else "fas fa-folder"),
                    "hide": mi.hide
                }

                if parent != mi.id:
                    # this is a sub item
                    menu[parent]["items"].append(_mi)
                    if _cnt:
                        #add its count to the header count
                        menu[parent]["count"] += _cnt
                else:
                    if len(menu[parent]["items"]):
                        # this is a top folder dont update the count.
                        del(_mi["count"])
                    menu[parent].update(_mi)
            except Exception as e:
                logger.exception(e)

        # reset to list
        menu = list(menu.values())

        # sort the menu list as the parents may be out of order.
        menu.sort(key=lambda i: i['rank'])

        # ensure no empty groups
        menu = filter(cls.filter_items, menu)

        return menu



import logging
from allianceauth.hooks import get_hooks

from django_registration.backends.activation.views import (
    REGISTRATION_SALT, ActivationView as BaseActivationView,
    RegistrationView as BaseRegistrationView,
)
from django_registration.signals import user_registered

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import authenticate, login
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.core import signing
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.template.loader import render_to_string
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _

from esi.decorators import token_required
from esi.models import Token

from allianceauth.eveonline.models import EveCharacter

from .core.celery_workers import active_tasks_count, queued_tasks_count
from .forms import RegistrationForm
from .models import CharacterOwnership

if 'allianceauth.eveonline.autogroups' in settings.INSTALLED_APPS:
    _has_auto_groups = True
    from allianceauth.eveonline.autogroups.models import *
else:
    _has_auto_groups = False


logger = logging.getLogger(__name__)


@login_required
def index(request):
    return redirect('authentication:dashboard')


def dashboard_groups(request):
    groups = request.user.groups.all()
    if _has_auto_groups:
        groups = groups\
            .filter(managedalliancegroup__isnull=True)\
            .filter(managedcorpgroup__isnull=True)
    groups = groups.order_by('name')

    context = {
        'groups': groups,
    }
    return render_to_string('authentication/dashboard.groups.html', context=context, request=request)


def dashboard_characters(request):
    characters = EveCharacter.objects\
        .filter(character_ownership__user=request.user)\
        .select_related()\
        .order_by('character_name')

    context = {
        'characters': characters
    }
    return render_to_string('authentication/dashboard.characters.html', context=context, request=request)


def dashboard_admin(request):
    if request.user.is_superuser:
        return render_to_string('allianceauth/admin-status/include.html', request=request)
    else:
        return ""


@login_required
def dashboard(request):
    _dash_items = list()
    hooks = get_hooks('dashboard_hook')
    items = [fn() for fn in hooks]
    items.sort(key=lambda i: i.order)
    for item in items:
        _dash_items.append(item.render(request))

    context = {
        'views': _dash_items,
    }
    return render(request, 'authentication/dashboard.html', context)


@login_required
def token_management(request):
    tokens = request.user.token_set.all()

    context = {
        'tokens': tokens
    }
    return render(request, 'authentication/tokens.html', context)


@login_required
def token_delete(request, token_id=None):
    try:
        token = Token.objects.get(id=token_id)
        if request.user == token.user:
            token.delete()
            messages.success(request, "Token Deleted.")
        else:
            messages.error(request, "This token does not belong to you.")
    except Token.DoesNotExist:
        messages.warning(request, "Token does not exist")
    return redirect('authentication:token_management')


@login_required
def token_refresh(request, token_id=None):
    try:
        token = Token.objects.get(id=token_id)
        if request.user == token.user:
            try:
                token.refresh()
                messages.success(request, "Token refreshed.")
            except Exception as e:
                messages.warning(request, f"Failed to refresh token. {e}")
        else:
            messages.error(request, "This token does not belong to you.")
    except Token.DoesNotExist:
        messages.warning(request, "Token does not exist")
    return redirect('authentication:token_management')


@login_required
@token_required(scopes=settings.LOGIN_TOKEN_SCOPES)
def main_character_change(request, token):
    logger.debug(f"main_character_change called by user {request.user} for character {token.character_name}")
    try:
        co = CharacterOwnership.objects.get(character__character_id=token.character_id, user=request.user)
    except CharacterOwnership.DoesNotExist:
        if not CharacterOwnership.objects.filter(character__character_id=token.character_id).exists():
            co = CharacterOwnership.objects.create_by_token(token)
        else:
            messages.error(
                request,
                _('Cannot change main character to %(char)s: character owned by a different account.') % ({'char': token.character_name})
            )
            co = None
    if co:
        request.user.profile.main_character = co.character
        request.user.profile.save(update_fields=['main_character'])
        messages.success(request, _('Changed main character to %(char)s') % {"char": co.character})
        logger.info('Changed user %(user)s main character to %(char)s' % ({'user': request.user, 'char': co.character}))
    return redirect("authentication:dashboard")


@token_required(new=True, scopes=settings.LOGIN_TOKEN_SCOPES)
def add_character(request, token):
    if CharacterOwnership.objects.filter(character__character_id=token.character_id).filter(
            owner_hash=token.character_owner_hash).filter(user=request.user).exists():
        messages.success(request, _('Added %(name)s to your account.' % ({'name': token.character_name})))
    else:
        messages.error(request, _('Failed to add %(name)s to your account: they already have an account.' % ({'name': token.character_name})))
    return redirect('authentication:dashboard')


"""
Override the HMAC two-step registration view to accommodate the three-step registration required.
Step 1: OAuth token to create user and profile.
Step 2: Get email and send activation link (but do not save email).
Step 3: Get link, save email and activate.

Step 1 is necessary to automatically assign character ownership and a main character, both of which require a saved User
model - this means the ensuing registration form cannot create the user because it already exists.

Email is not saved to the user model in Step 2 as a way of differentiating users who have not yet completed registration
(is_active=False) and users who have been disabled by an admin (is_active=False, email present).

Because of this, the email address needs to be assigned in Step 3 after clicking the link, which means the link must
have the email address embedded much like the username. Key creation and decoding is overridden to support this action.
"""


# Step 1
@token_required(new=True, scopes=settings.LOGIN_TOKEN_SCOPES)
def sso_login(request, token):
    user = authenticate(token=token)
    if user:
        token.user = user
        if Token.objects.exclude(pk=token.pk).equivalent_to(token).require_valid().exists():
            token.delete()
        else:
            token.save()
        if user.is_active:
            login(request, user)
            return redirect(request.POST.get('next', request.GET.get('next', 'authentication:dashboard')))
        elif not user.email:
            # Store the new user PK in the session to enable us to identify the registering user in Step 2
            request.session['registration_uid'] = user.pk
            # Go to Step 2
            return redirect('registration_register')
    messages.error(request, _('Unable to authenticate as the selected character.'))
    return redirect(settings.LOGIN_URL)


# Step 2
class RegistrationView(BaseRegistrationView):
    form_class = RegistrationForm
    template_name = "public/register.html"
    email_body_template = "registration/activation_email.txt"
    email_body_template_html = "registration/activation_email_html.txt"
    email_subject_template = "registration/activation_email_subject.txt"
    success_url = reverse_lazy('registration_complete')

    def send_activation_email(self, user):
        """
        Implement our own way to send a mail to make sure we
        send a RFC conform multipart email
        :param user:
        :type user:
        """

        activation_key = self.get_activation_key(user)
        context = self.get_email_context(activation_key)
        context["user"] = user

        # email subject
        subject = render_to_string(
            template_name=self.email_subject_template,
            context=context,
            request=self.request,
        )
        subject = "".join(subject.splitlines())

        # plaintext email body part
        message = render_to_string(
            template_name=self.email_body_template,
            context=context,
            request=self.request,
        )

        # html email body part
        message_html = render_to_string(
            template_name=self.email_body_template_html,
            context=context,
            request=self.request,
        )

        # send it
        user.email_user(
            subject,
            message,
            settings.DEFAULT_FROM_EMAIL,
            **{'html_message': message_html},
        )

    def get_success_url(self, user):
        if not getattr(settings, 'REGISTRATION_VERIFY_EMAIL', True):
            return reverse_lazy('authentication:dashboard')
        return super().get_success_url(user)

    def dispatch(self, request, *args, **kwargs):
        # We're storing a key in the session to pass user information from OAuth response. Make sure it's there.
        if not self.request.session.get('registration_uid', None) or not User.objects.filter(
                pk=self.request.session.get('registration_uid')).exists():
            messages.error(self.request, _('Registration token has expired.'))
            return redirect(settings.LOGIN_URL)
        if not getattr(settings, 'REGISTRATION_VERIFY_EMAIL', True):
            # Keep the request so the user can be automagically logged in.
            setattr(self, 'request', request)
        return super().dispatch(request, *args, **kwargs)

    def register(self, form):
        user = User.objects.get(pk=self.request.session.get('registration_uid'))
        user.email = form.cleaned_data['email']
        user_registered.send(self.__class__, user=user, request=self.request)
        if getattr(settings, 'REGISTRATION_VERIFY_EMAIL', True):
            # Go to Step 3
            self.send_activation_email(user)
        else:
            user.is_active = True
            user.save()
            login(self.request, user, 'allianceauth.authentication.backends.StateBackend')
        return user

    def get_activation_key(self, user):
        return signing.dumps(obj=[getattr(user, User.USERNAME_FIELD), user.email], salt=REGISTRATION_SALT)

    def get_email_context(self, activation_key):
        context = super().get_email_context(activation_key)
        context['url'] = context['site'].domain + reverse('registration_activate', args=[activation_key])
        return context


# Step 3
class ActivationView(BaseActivationView):
    template_name = "registration/activate.html"
    success_url = reverse_lazy('registration_activation_complete')

    def validate_key(self, activation_key):
        try:
            dump = signing.loads(
                activation_key,
                salt=REGISTRATION_SALT,
                max_age=settings.ACCOUNT_ACTIVATION_DAYS * 86400
            )
            return dump
        except signing.BadSignature:
            return None

    def activate(self, *args, **kwargs):
        dump = self.validate_key(kwargs.get('activation_key'))
        if dump:
            user = self.get_user(dump[0])
            if user:
                user.email = dump[1]
                user.is_active = True
                user.save()
                return user
        return False


def registration_complete(request):
    messages.success(request, _('Sent confirmation email. Please follow the link to confirm your email address.'))
    return redirect('authentication:login')


def activation_complete(request):
    messages.success(request, _('Confirmed your email address. Please login to continue.'))
    return redirect('authentication:dashboard')


def registration_closed(request):
    messages.error(request, _('Registration of new accounts is not allowed at this time.'))
    return redirect('authentication:login')


def task_counts(request) -> JsonResponse:
    """Return task counts as JSON for an AJAX call."""
    data = {
        "tasks_running": active_tasks_count(),
        "tasks_queued": queued_tasks_count()
    }
    return JsonResponse(data)



import logging

from celery import shared_task

from .models import EveAllianceInfo, EveCharacter, EveCorporationInfo
from . import providers


logger = logging.getLogger(__name__)

TASK_PRIORITY = 7
CHUNK_SIZE = 500


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


@shared_task
def update_corp(corp_id):
    """Update given corporation from ESI"""
    EveCorporationInfo.objects.update_corporation(corp_id)


@shared_task
def update_alliance(alliance_id):
    """Update given alliance from ESI"""
    EveAllianceInfo.objects.update_alliance(alliance_id).populate_alliance()


@shared_task
def update_character(character_id: int) -> None:
    """Update given character from ESI."""
    EveCharacter.objects.update_character(character_id)


@shared_task
def run_model_update():
    """Update all alliances, corporations and characters from ESI"""

    #update existing corp models
    for corp in EveCorporationInfo.objects.all().values('corporation_id'):
        update_corp.apply_async(args=[corp['corporation_id']], priority=TASK_PRIORITY)

    # update existing alliance models
    for alliance in EveAllianceInfo.objects.all().values('alliance_id'):
        update_alliance.apply_async(args=[alliance['alliance_id']], priority=TASK_PRIORITY)

    # update existing character models
    character_ids = EveCharacter.objects.all().values_list('character_id', flat=True)
    for character_ids_chunk in chunks(character_ids, CHUNK_SIZE):
        update_character_chunk.apply_async(
            args=[character_ids_chunk], priority=TASK_PRIORITY
        )


@shared_task
def update_character_chunk(character_ids_chunk: list):
    """Update a list of character from ESI"""
    try:
        affiliations_raw = providers.provider.client.Character\
            .post_characters_affiliation(characters=character_ids_chunk).result()
        character_names = providers.provider.client.Universe\
            .post_universe_names(ids=character_ids_chunk).result()
    except OSError:
        logger.info("Failed to bulk update characters. Attempting single updates")
        for character_id in character_ids_chunk:
            update_character.apply_async(
                args=[character_id], priority=TASK_PRIORITY
            )
        return

    affiliations = {
        affiliation.get('character_id'): affiliation
        for affiliation in affiliations_raw
    }
    # add character names to affiliations
    for character in character_names:
        character_id = character.get('id')
        if character_id in affiliations:
            affiliations[character_id]['name'] = character.get('name')

    # fetch current characters
    characters = EveCharacter.objects.filter(character_id__in=character_ids_chunk)\
        .values('character_id', 'corporation_id', 'alliance_id', 'character_name')

    for character in characters:
        character_id = character.get('character_id')
        if character_id in affiliations:
            affiliation = affiliations[character_id]

            corp_changed = (
                character.get('corporation_id') != affiliation.get('corporation_id')
            )

            alliance_id = character.get('alliance_id')
            if not alliance_id:
                alliance_id = None
            alliance_changed = alliance_id != affiliation.get('alliance_id')

            name_changed = False
            fetched_name = affiliation.get('name', False)
            if fetched_name:
                name_changed = character.get('character_name') != fetched_name

            if corp_changed or alliance_changed or name_changed:
                update_character.apply_async(
                    args=[character.get('character_id')], priority=TASK_PRIORITY
                )



from unittest import mock
from allianceauth.authentication.middleware import UserSettingsMiddleware
from unittest.mock import Mock
from django.http import HttpResponse

from django.test.testcases import TestCase


class TestUserSettingsMiddlewareSaveLang(TestCase):

    def setUp(self):
        self.middleware = UserSettingsMiddleware(HttpResponse)
        self.request = Mock()
        self.request.headers = {
                "User-Agent": "AUTOMATED TEST"
            }
        self.request.path = '/i18n/setlang/'
        self.request.POST = {
            'language': 'fr'
        }
        self.request.user.profile.language = 'de'
        self.request.user.is_anonymous = False
        self.response = Mock()
        self.response.content = 'hello world'

    def test_middleware_passthrough(self):
        """
        Simply tests the middleware runs cleanly
        """
        response = self.middleware.process_response(
            self.request,
            self.response
            )
        self.assertEqual(self.response, response)

    def test_middleware_save_language_false_anonymous(self):
        """
        Ensures the middleware wont change the usersettings
        of a non-existent (anonymous) user
        """
        self.request.user.is_anonymous = True
        response = self.middleware.process_response(
            self.request,
            self.response
            )
        self.assertEqual(self.request.user.profile.language, 'de')
        self.assertFalse(self.request.user.profile.save.called)
        self.assertEqual(self.request.user.profile.save.call_count, 0)

    def test_middleware_save_language_new(self):
        """
        does the middleware change a language not set in the DB
        """
        self.request.user.profile.language = None
        response = self.middleware.process_response(
            self.request,
            self.response
            )
        self.assertEqual(self.request.user.profile.language, 'fr')
        self.assertTrue(self.request.user.profile.save.called)
        self.assertEqual(self.request.user.profile.save.call_count, 1)

    def test_middleware_save_language_changed(self):
        """
        Tests the middleware will change a language setting
        """
        response = self.middleware.process_response(
            self.request,
            self.response
            )
        self.assertEqual(self.request.user.profile.language, 'fr')
        self.assertTrue(self.request.user.profile.save.called)
        self.assertEqual(self.request.user.profile.save.call_count, 1)


class TestUserSettingsMiddlewareLoginFlow(TestCase):

    def setUp(self):
        self.middleware = UserSettingsMiddleware(HttpResponse)
        self.request = Mock()
        self.request.headers = {
                "User-Agent": "AUTOMATED TEST"
            }
        self.request.path = '/sso/login'
        self.request.session = {
            'NIGHT_MODE': False
        }
        self.request.LANGUAGE_CODE = 'en'
        self.request.user.profile.language = 'de'
        self.request.user.profile.night_mode = True
        self.request.user.is_anonymous = False
        self.response = Mock()
        self.response.content = 'hello world'

    def test_middleware_passthrough(self):
        """
        Simply tests the middleware runs cleanly
        """
        middleware_response = self.middleware.process_response(
            self.request,
            self.response
            )
        self.assertEqual(self.response, middleware_response)

    def test_middleware_sets_language_cookie_true_no_cookie(self):
        """
        tests the middleware will set a cookie, while none is set
        """
        self.request.LANGUAGE_CODE = None
        middleware_response = self.middleware.process_response(
            self.request,
            self.response
            )
        self.assertTrue(middleware_response.set_cookie.called)
        self.assertEqual(middleware_response.set_cookie.call_count, 1)
        args, kwargs = middleware_response.set_cookie.call_args
        self.assertEqual(kwargs['value'], 'de')

    def test_middleware_sets_language_cookie_true_wrong_cookie(self):
        """
        tests the middleware will set a cookie, while a different value is set
        """
        middleware_response = self.middleware.process_response(
            self.request,
            self.response
            )
        self.assertTrue(middleware_response.set_cookie.called)
        self.assertEqual(middleware_response.set_cookie.call_count, 1)
        args, kwargs = middleware_response.set_cookie.call_args
        self.assertEqual(kwargs['value'], 'de')

    def test_middleware_sets_language_cookie_false_anonymous(self):
        """
        ensures the middleware wont set a value for a non existent user (anonymous)
        """
        self.request.user.is_anonymous = True
        middleware_response = self.middleware.process_response(
            self.request,
            self.response
            )
        self.assertFalse = middleware_response.set_cookie.called
        self.assertEqual(middleware_response.set_cookie.call_count, 0)

    def test_middleware_sets_language_cookie_false_already_set(self):
        """
        tests the middleware skips setting the cookie, if its already set correctly
        """
        self.request.user.profile.language = 'en'
        middleware_response = self.middleware.process_response(
            self.request,
            self.response
            )
        self.assertFalse = middleware_response.set_cookie.called
        self.assertEqual(middleware_response.set_cookie.call_count, 0)

    def test_middleware_sets_night_mode_not_set(self):
        """
        tests the middleware will set night_mode if not set
        """
        self.request.session = {}
        response = self.middleware.process_response(
            self.request,
            self.response
            )
        self.assertEqual(self.request.session["NIGHT_MODE"], True)

    def test_middleware_sets_night_mode_set(self):
        """
        tests the middleware will set night_mode if set.
        """
        response = self.middleware.process_response(
            self.request,
            self.response
            )
        self.assertEqual(self.request.session["NIGHT_MODE"], True)



from unittest.mock import patch

from django_webtest import WebTest

from django.conf import settings
from django.contrib import admin
from django.contrib.admin.sites import AdminSite
from django.contrib.auth.models import User
from django.test import TestCase, RequestFactory, Client, override_settings

from allianceauth.authentication.models import CharacterOwnership, State
from allianceauth.eveonline.models import (
    EveCharacter, EveCorporationInfo, EveAllianceInfo
)
from allianceauth.tests.auth_utils import AuthUtils

from . import get_admin_change_view_url
from ..admin import HasLeaderFilter, GroupAdmin, Group
from ..models import ReservedGroupName


if 'allianceauth.eveonline.autogroups' in settings.INSTALLED_APPS:
    _has_auto_groups = True
    from allianceauth.eveonline.autogroups.models import AutogroupsConfig
    from ..admin import IsAutoGroupFilter
else:
    _has_auto_groups = False


MODULE_PATH = 'allianceauth.groupmanagement.admin'


class MockRequest:

    def __init__(self, user=None):
        self.user = user


class TestGroupAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # group 1 - has leader
        cls.group_1 = Group.objects.create(name='Group 1')
        cls.group_1.authgroup.description = 'Default Group'
        cls.group_1.authgroup.internal = False
        cls.group_1.authgroup.hidden = False
        cls.group_1.authgroup.save()

        # group 2 - no leader
        cls.group_2 = Group.objects.create(name='Group 2')
        cls.group_2.authgroup.description = 'Internal Group'
        cls.group_2.authgroup.internal = True
        cls.group_2.authgroup.group_leader_groups.add(cls.group_1)
        cls.group_2.authgroup.save()

        # group 3 - has leader
        cls.group_3 = Group.objects.create(name='Group 3')
        cls.group_3.authgroup.description = 'Hidden Group'
        cls.group_3.authgroup.internal = False
        cls.group_3.authgroup.hidden = True
        cls.group_3.authgroup.save()

        # group 4 - no leader
        cls.group_4 = Group.objects.create(name='Group 4')
        cls.group_4.authgroup.description = 'Open Group'
        cls.group_4.authgroup.internal = False
        cls.group_4.authgroup.hidden = False
        cls.group_4.authgroup.open = True
        cls.group_4.authgroup.save()

        # group 5 - no leader
        cls.group_5 = Group.objects.create(name='Group 5')
        cls.group_5.authgroup.description = 'Public Group'
        cls.group_5.authgroup.internal = False
        cls.group_5.authgroup.hidden = False
        cls.group_5.authgroup.public = True
        cls.group_5.authgroup.save()

        # group 6 - no leader
        cls.group_6 = Group.objects.create(name='Group 6')
        cls.group_6.authgroup.description = 'Mixed Group'
        cls.group_6.authgroup.internal = False
        cls.group_6.authgroup.hidden = True
        cls.group_6.authgroup.open = True
        cls.group_6.authgroup.public = True
        cls.group_6.authgroup.save()

        # user 1 - corp and alliance, normal user
        cls.character_1 = EveCharacter.objects.create(
            character_id=1001,
            character_name='Bruce Wayne',
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',
        )
        cls.character_1a = EveCharacter.objects.create(
            character_id=1002,
            character_name='Batman',
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',
        )
        alliance = EveAllianceInfo.objects.create(
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',
            executor_corp_id=2001
        )
        EveCorporationInfo.objects.create(
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',
            member_count=42,
            alliance=alliance
        )
        cls.user_1 = User.objects.create_user(
            cls.character_1.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=cls.character_1,
            owner_hash='x1' + cls.character_1.character_name,
            user=cls.user_1
        )
        CharacterOwnership.objects.create(
            character=cls.character_1a,
            owner_hash='x1' + cls.character_1a.character_name,
            user=cls.user_1
        )
        cls.user_1.profile.main_character = cls.character_1
        cls.user_1.profile.save()
        cls.user_1.groups.add(cls.group_1)
        cls.group_1.authgroup.group_leaders.add(cls.user_1)

        # user 2 - corp only, staff
        cls.character_2 = EveCharacter.objects.create(
            character_id=1003,
            character_name='Clark Kent',
            corporation_id=2002,
            corporation_name='Daily Planet',
            corporation_ticker='DP',
            alliance_id=None
        )
        EveCorporationInfo.objects.create(
            corporation_id=2002,
            corporation_name='Daily Plane',
            corporation_ticker='DP',
            member_count=99,
            alliance=None
        )
        cls.user_2 = User.objects.create_user(
            cls.character_2.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=cls.character_2,
            owner_hash='x1' + cls.character_2.character_name,
            user=cls.user_2
        )
        cls.user_2.profile.main_character = cls.character_2
        cls.user_2.profile.save()
        cls.user_2.groups.add(cls.group_2)
        cls.user_2.is_staff = True
        cls.user_2.save()

        # user 3 - no main, no group, superuser
        cls.character_3 = EveCharacter.objects.create(
            character_id=1101,
            character_name='Lex Luthor',
            corporation_id=2101,
            corporation_name='Lex Corp',
            corporation_ticker='LC',
            alliance_id=None
        )
        EveCorporationInfo.objects.create(
            corporation_id=2101,
            corporation_name='Lex Corp',
            corporation_ticker='LC',
            member_count=666,
            alliance=None
        )
        EveAllianceInfo.objects.create(
            alliance_id=3101,
            alliance_name='Lex World Domination',
            alliance_ticker='LWD',
            executor_corp_id=2101
        )
        cls.user_3 = User.objects.create_user(
            cls.character_3.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=cls.character_3,
            owner_hash='x1' + cls.character_3.character_name,
            user=cls.user_3
        )
        cls.user_3.is_superuser = True
        cls.user_3.save()
        cls.user_3.groups.add(cls.group_3)
        cls.group_3.authgroup.group_leaders.add(cls.user_3)

    def setUp(self):
        self.factory = RequestFactory()
        self.modeladmin = GroupAdmin(
            model=Group, admin_site=AdminSite()
        )

    def _create_autogroups(self):
        """create autogroups for corps and alliances"""
        if _has_auto_groups:
            autogroups_config = AutogroupsConfig(
                corp_groups=True,
                alliance_groups=True
            )
            autogroups_config.save()
            for state in State.objects.all():
                autogroups_config.states.add(state)
            autogroups_config.update_corp_group_membership(self.user_1)

    # column rendering

    def test_description(self):
        expected = 'Default Group'
        result = self.modeladmin._description(self.group_1)
        self.assertEqual(result, expected)

    def test_member_count(self):
        # given
        request = MockRequest(user=self.user_1)
        obj = self.modeladmin.get_queryset(request).get(pk=self.group_1.pk)
        # when
        result = self.modeladmin._member_count(obj)
        # then
        self.assertEqual(result, 1)

    def test_has_leader_user(self):
        # given
        request = MockRequest(user=self.user_1)
        obj = self.modeladmin.get_queryset(request).get(pk=self.group_1.pk)
        # when
        result = self.modeladmin.has_leader(obj)
        # then
        self.assertTrue(result)

    def test_has_leader_group(self):
        # given
        request = MockRequest(user=self.user_1)
        obj = self.modeladmin.get_queryset(request).get(pk=self.group_2.pk)
        # when
        result = self.modeladmin.has_leader(obj)
        # then
        self.assertTrue(result)

    def test_properties_1(self):
        # given
        request = MockRequest(user=self.user_1)
        obj = self.modeladmin.get_queryset(request).get(pk=self.group_1.pk)
        # when
        result = self.modeladmin._properties(obj)
        self.assertListEqual(result, ['Default'])

    def test_properties_2(self):
        # given
        request = MockRequest(user=self.user_1)
        obj = self.modeladmin.get_queryset(request).get(pk=self.group_2.pk)
        # when
        result = self.modeladmin._properties(obj)
        self.assertListEqual(result, ['Internal'])

    def test_properties_3(self):
        # given
        request = MockRequest(user=self.user_1)
        obj = self.modeladmin.get_queryset(request).get(pk=self.group_3.pk)
        # when
        result = self.modeladmin._properties(obj)
        self.assertListEqual(result, ['Hidden'])

    def test_properties_4(self):
        # given
        request = MockRequest(user=self.user_1)
        obj = self.modeladmin.get_queryset(request).get(pk=self.group_4.pk)
        # when
        result = self.modeladmin._properties(obj)
        self.assertListEqual(result, ['Open'])

    def test_properties_5(self):
        # given
        request = MockRequest(user=self.user_1)
        obj = self.modeladmin.get_queryset(request).get(pk=self.group_5.pk)
        # when
        result = self.modeladmin._properties(obj)
        self.assertListEqual(result, ['Public'])

    def test_properties_6(self):
        # given
        request = MockRequest(user=self.user_1)
        obj = self.modeladmin.get_queryset(request).get(pk=self.group_6.pk)
        # when
        result = self.modeladmin._properties(obj)
        self.assertListEqual(result, ['Hidden', 'Open', 'Public'])

    if _has_auto_groups:
        @patch(MODULE_PATH + '._has_auto_groups', True)
        def test_should_show_autogroup_for_corporation(self):
            # given
            self._create_autogroups()
            request = MockRequest(user=self.user_1)
            queryset = self.modeladmin.get_queryset(request)
            obj = queryset.filter(managedcorpgroup__isnull=False).first()
            # when
            result = self.modeladmin._properties(obj)
            # then
            self.assertListEqual(result, ['Auto Group'])

        @patch(MODULE_PATH + '._has_auto_groups', True)
        def test_should_show_autogroup_for_alliance(self):
            # given
            self._create_autogroups()
            request = MockRequest(user=self.user_1)
            queryset = self.modeladmin.get_queryset(request)
            obj = queryset.filter(managedalliancegroup__isnull=False).first()
            # when
            result = self.modeladmin._properties(obj)
            # then
            self.assertListEqual(result, ['Auto Group'])

    # actions

    # filters

    if _has_auto_groups:
        @patch(MODULE_PATH + '._has_auto_groups', True)
        def test_filter_is_auto_group(self):

            class GroupAdminTest(admin.ModelAdmin):
                list_filter = (IsAutoGroupFilter,)

            self._create_autogroups()
            my_modeladmin = GroupAdminTest(Group, AdminSite())

            # Make sure the lookups are correct
            request = self.factory.get('/')
            request.user = self.user_1
            changelist = my_modeladmin.get_changelist_instance(request)
            filters = changelist.get_filters(request)
            filterspec = filters[0][0]
            expected = [
                ('yes', 'Yes'),
                ('no', 'No'),
            ]
            self.assertEqual(filterspec.lookup_choices, expected)

            # Make sure the correct queryset is returned - no
            request = self.factory.get(
                '/', {'is_auto_group__exact': 'no'}
            )
            request.user = self.user_1
            changelist = my_modeladmin.get_changelist_instance(request)
            queryset = changelist.get_queryset(request)
            expected = [
                self.group_1,
                self.group_2,
                self.group_3,
                self.group_4,
                self.group_5,
                self.group_6
            ]
            self.assertSetEqual(set(queryset), set(expected))

            # Make sure the correct queryset is returned - yes
            request = self.factory.get(
                '/', {'is_auto_group__exact': 'yes'}
            )
            request.user = self.user_1
            changelist = my_modeladmin.get_changelist_instance(request)
            queryset = changelist.get_queryset(request)
            expected = Group.objects.exclude(
                managedalliancegroup__isnull=True,
                managedcorpgroup__isnull=True
            )
            self.assertSetEqual(set(queryset), set(expected))

    def test_filter_has_leader(self):

        class GroupAdminTest(admin.ModelAdmin):
            list_filter = (HasLeaderFilter,)

        self._create_autogroups()
        my_modeladmin = GroupAdminTest(Group, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get('/')
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        filters = changelist.get_filters(request)
        filterspec = filters[0][0]
        expected = [
            ('yes', 'Yes'),
            ('no', 'No'),
        ]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned - no
        request = self.factory.get(
            '/', {'has_leader__exact': 'no'}
        )
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = Group.objects.exclude(pk__in=[
            self.group_1.pk, self.group_3.pk
        ])
        self.assertSetEqual(set(queryset), set(expected))

        # Make sure the correct queryset is returned - yes
        request = self.factory.get(
            '/', {'has_leader__exact': 'yes'}
        )
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = [
            self.group_1,
            self.group_3
        ]
        self.assertSetEqual(set(queryset), set(expected))

    def test_change_view_loads_normally(self):
        User.objects.create_superuser(
            username='superuser', password='secret', email='admin@example.com'
        )
        c = Client()
        c.login(username='superuser', password='secret')
        response = c.get(get_admin_change_view_url(self.group_1))
        self.assertEqual(response.status_code, 200)

    def test_should_create_new_group(self):
        # given
        user = User.objects.create_superuser("bruce")
        self.client.force_login(user)
        # when
        response = self.client.post(
            "/admin/groupmanagement/group/add/",
            data={
                "name": "new group",
                "authgroup-TOTAL_FORMS": 1,
                "authgroup-INITIAL_FORMS": 0,
                "authgroup-MIN_NUM_FORMS": 0,
                "authgroup-MAX_NUM_FORMS": 1,
            }
        )
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/groupmanagement/group/")
        self.assertTrue(Group.objects.filter(name="new group").exists())

    def test_should_not_allow_creating_new_group_with_reserved_name(self):
        # given
        ReservedGroupName.objects.create(
            name="new group", reason="dummy", created_by="bruce"
        )
        user = User.objects.create_superuser("bruce")
        self.client.force_login(user)
        # when
        response = self.client.post(
            "/admin/groupmanagement/group/add/",
            data={
                "name": "New group",
                "authgroup-TOTAL_FORMS": 1,
                "authgroup-INITIAL_FORMS": 0,
                "authgroup-MIN_NUM_FORMS": 0,
                "authgroup-MAX_NUM_FORMS": 1,
            }
        )
        # then
        self.assertContains(
            response, "This name has been reserved and can not be used for groups"
        )
        self.assertFalse(Group.objects.filter(name="new group").exists())

    def test_should_not_allow_changing_name_of_existing_group_to_reserved_name(self):
        # given
        ReservedGroupName.objects.create(
            name="new group", reason="dummy", created_by="bruce"
        )
        group = Group.objects.create(name="dummy")
        user = User.objects.create_superuser("bruce")
        self.client.force_login(user)
        # when
        response = self.client.post(
            f"/admin/groupmanagement/group/{group.pk}/change/",
            data={
                "name": "new group",
                "authgroup-TOTAL_FORMS": 1,
                "authgroup-INITIAL_FORMS": 0,
                "authgroup-MIN_NUM_FORMS": 0,
                "authgroup-MAX_NUM_FORMS": 1,
            }
        )
        # then
        self.assertContains(
            response, "This name has been reserved and can not be used for groups"
        )
        self.assertFalse(Group.objects.filter(name="new group").exists())


class TestGroupAdminChangeFormSuperuserExclusiveEdits(WebTest):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.super_admin = User.objects.create_superuser("super_admin")
        cls.staff_admin = User.objects.create_user("staff_admin")
        cls.staff_admin.is_staff = True
        cls.staff_admin.save()
        cls.staff_admin = AuthUtils.add_permissions_to_user_by_name(
            [
                "auth.add_group",
                "auth.change_group",
                "auth.view_group",
                "groupmanagement.add_group",
                "groupmanagement.change_group",
                "groupmanagement.view_group",
            ],
            cls.staff_admin
        )
        cls.superuser_exclusive_fields = ["permissions", "authgroup-0-restricted"]

    def test_should_show_all_fields_to_superuser_for_add(self):
        # given
        self.app.set_user(self.super_admin)
        page = self.app.get("/admin/groupmanagement/group/add/")
        # when
        form = page.forms["group_form"]
        # then
        for field in self.superuser_exclusive_fields:
            with self.subTest(field=field):
                self.assertIn(field, form.fields)

    def test_should_not_show_all_fields_to_staff_admins_for_add(self):
        # given
        self.app.set_user(self.staff_admin)
        page = self.app.get("/admin/groupmanagement/group/add/")
        # when
        form = page.forms["group_form"]
        # then
        for field in self.superuser_exclusive_fields:
            with self.subTest(field=field):
                self.assertNotIn(field, form.fields)

    def test_should_show_all_fields_to_superuser_for_change(self):
        # given
        self.app.set_user(self.super_admin)
        group = Group.objects.create(name="Dummy group")
        page = self.app.get(f"/admin/groupmanagement/group/{group.pk}/change/")
        # when
        form = page.forms["group_form"]
        # then
        for field in self.superuser_exclusive_fields:
            with self.subTest(field=field):
                self.assertIn(field, form.fields)

    def test_should_not_show_all_fields_to_staff_admin_for_change(self):
        # given
        self.app.set_user(self.staff_admin)
        group = Group.objects.create(name="Dummy group")
        page = self.app.get(f"/admin/groupmanagement/group/{group.pk}/change/")
        # when
        form = page.forms["group_form"]
        # then
        for field in self.superuser_exclusive_fields:
            with self.subTest(field=field):
                self.assertNotIn(field, form.fields)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestGroupAdmin2(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.superuser = User.objects.create_superuser("super")

    def test_should_remove_users_from_state_groups(self):
        # given
        user_member = AuthUtils.create_user("Bruce Wayne")
        character_member = AuthUtils.add_main_character_2(
            user_member,
            name="Bruce Wayne",
            character_id=1001,
            corp_id=2001,
            corp_name="Wayne Technologies",
        )
        user_guest = AuthUtils.create_user("Lex Luthor")
        AuthUtils.add_main_character_2(
            user_guest,
            name="Lex Luthor",
            character_id=1011,
            corp_id=2011,
            corp_name="Luthor Corp",
        )
        member_state = AuthUtils.get_member_state()
        member_state.member_characters.add(character_member)
        user_member.refresh_from_db()
        user_guest.refresh_from_db()
        group = Group.objects.create(name="dummy")
        user_member.groups.add(group)
        user_guest.groups.add(group)
        group.authgroup.states.add(member_state)
        self.client.force_login(self.superuser)
        # when
        response = self.client.post(
            f"/admin/groupmanagement/group/{group.pk}/change/",
            data={
                "name": f"{group.name}",
                "authgroup-TOTAL_FORMS": "1",
                "authgroup-INITIAL_FORMS": "1",
                "authgroup-MIN_NUM_FORMS": "0",
                "authgroup-MAX_NUM_FORMS": "1",
                "authgroup-0-description": "",
                "authgroup-0-states": f"{member_state.pk}",
                "authgroup-0-internal": "on",
                "authgroup-0-hidden": "on",
                "authgroup-0-group": f"{group.pk}",
                "authgroup-__prefix__-description": "",
                "authgroup-__prefix__-internal": "on",
                "authgroup-__prefix__-hidden": "on",
                "authgroup-__prefix__-group": f"{group.pk}",
                "_save": "Save"
            }
        )
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/groupmanagement/group/")
        self.assertIn(group, user_member.groups.all())
        self.assertNotIn(group, user_guest.groups.all())


class TestReservedGroupNameAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_superuser("bruce")

    def test_should_create_new_entry(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            "/admin/groupmanagement/reservedgroupname/add/",
            data={"name": "Test", "reason": "dummy"}
        )
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/groupmanagement/reservedgroupname/")
        obj = ReservedGroupName.objects.get(name="test")
        self.assertEqual(obj.name, "test")
        self.assertEqual(obj.created_by, self.user.username)
        self.assertTrue(obj.created_at)

    def test_should_not_allow_names_of_existing_groups(self):
        # given
        Group.objects.create(name="Already taken")
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            "/admin/groupmanagement/reservedgroupname/add/",
            data={"name": "already taken", "reason": "dummy"}
        )
        # then
        self.assertContains(response, "There already exists a group with that name")
        self.assertFalse(ReservedGroupName.objects.filter(name="already taken").exists())



from unittest import mock

from celery_once import AlreadyQueued

from django.core.cache import cache
from django.test import override_settings, TestCase

from allianceauth.tests.auth_utils import AuthUtils
from allianceauth.services.tasks import validate_services, update_groups_for_user

from ..tasks import DjangoBackend


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class ServicesTasksTestCase(TestCase):
    def setUp(self):
        self.member = AuthUtils.create_user('auth_member')

    @mock.patch('allianceauth.services.tasks.ServicesHook')
    def test_validate_services(self, services_hook):
        # given
        svc = mock.Mock()
        svc.validate_user.return_value = None
        services_hook.get_services.return_value = [svc]
        # when
        validate_services.delay(self.member.pk)
        # then
        self.assertTrue(services_hook.get_services.called)
        self.assertTrue(svc.validate_user.called)
        args, _ = svc.validate_user.call_args
        self.assertEqual(self.member, args[0])  # Assert correct user is passed to service hook function

    @mock.patch('allianceauth.services.tasks.ServicesHook')
    def test_update_groups_for_user(self, services_hook):
        # given
        svc = mock.Mock()
        svc.validate_user.return_value = None
        services_hook.get_services.return_value = [svc]
        # when
        update_groups_for_user.delay(self.member.pk)
        # then
        self.assertTrue(services_hook.get_services.called)
        self.assertTrue(svc.validate_user.called)
        args, _ = svc.validate_user.call_args
        self.assertEqual(self.member, args[0])  # Assert correct user
        self.assertTrue(svc.update_groups.called)
        args, _ = svc.update_groups.call_args
        self.assertEqual(self.member, args[0])  # Assert correct user


class TestDjangoBackend(TestCase):

    TEST_KEY = "my-django-backend-test-key"
    TIMEOUT = 1800

    def setUp(self) -> None:
        cache.delete(self.TEST_KEY)
        self.backend = DjangoBackend(dict())

    def test_can_get_lock(self):
        """
        when lock can be acquired
        then set it with timetout
        """
        self.backend.raise_or_lock(self.TEST_KEY, self.TIMEOUT)
        self.assertIsNotNone(cache.get(self.TEST_KEY))
        self.assertAlmostEqual(cache.ttl(self.TEST_KEY), self.TIMEOUT, delta=2)

    def test_when_cant_get_lock_raise_exception(self):
        """
        when lock can bot be acquired
        then raise AlreadyQueued exception with countdown
        """
        self.backend.raise_or_lock(self.TEST_KEY, self.TIMEOUT)

        try:
            self.backend.raise_or_lock(self.TEST_KEY, self.TIMEOUT)
        except Exception as ex:
            self.assertIsInstance(ex, AlreadyQueued)
            self.assertAlmostEqual(ex.countdown, self.TIMEOUT, delta=2)

    def test_can_clear_lock(self):
        """
        when a lock exists
        then can get a new lock after clearing it
        """
        self.backend.raise_or_lock(self.TEST_KEY, self.TIMEOUT)

        self.backend.clear_lock(self.TEST_KEY)
        self.backend.raise_or_lock(self.TEST_KEY, self.TIMEOUT)
        self.assertIsNotNone(cache.get(self.TEST_KEY))



from django.contrib import admin

from . import models


@admin.register(models.MenuItem)
class MenuItemAdmin(admin.ModelAdmin):
    list_display = ['text', 'hide', 'parent', 'url', 'icon_classes', 'rank']
    ordering = ('rank',)



import logging

logger = logging.getLogger(__name__)


class NotificationHandler(logging.Handler):
    def emit(self, record):
        from django.contrib.auth.models import User, Permission
        from django.db.models import Q
        from . import notify
        from .models import Notification

        try:
            perm = Permission.objects.get(codename="logging_notifications")
        except Permission.DoesNotExist:
            return

        message = record.getMessage()
        if record.exc_text:
            message += "\n\n"
            message = message + record.exc_text

        users = User.objects.filter(
            Q(groups__permissions=perm) | Q(user_permissions=perm) | Q(is_superuser=True)).distinct()
        for user in users:
            notify(
                user,
                f"{record.levelname} [{record.funcName}:{record.lineno}]",
                level=Notification.Level.from_old_name(record.levelname),
                message=message
            )



# Generated by Django 1.10.1 on 2016-09-05 21:40

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('eveonline', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='SrpFleetMain',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('fleet_name', models.CharField(default=b'', max_length=254)),
                ('fleet_doctrine', models.CharField(default=b'', max_length=254)),
                ('fleet_time', models.DateTimeField()),
                ('fleet_srp_code', models.CharField(default=b'', max_length=254)),
                ('fleet_srp_status', models.CharField(default=b'', max_length=254)),
                ('fleet_srp_aar_link', models.CharField(default=b'', max_length=254)),
                ('fleet_commander', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveonline.EveCharacter')),
            ],
        ),
        migrations.CreateModel(
            name='SrpUserRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('killboard_link', models.CharField(default=b'', max_length=254)),
                ('after_action_report_link', models.CharField(default=b'', max_length=254)),
                ('additional_info', models.CharField(default=b'', max_length=254)),
                ('srp_status', models.CharField(default=b'', max_length=254)),
                ('srp_total_amount', models.BigIntegerField(default=0)),
                ('kb_total_loss', models.BigIntegerField(default=0)),
                ('srp_ship_name', models.CharField(default=b'', max_length=254)),
                ('post_time', models.DateTimeField(default=django.utils.timezone.now)),
                ('character', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveonline.EveCharacter')),
                ('srp_fleet_main', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='srp.SrpFleetMain')),
            ],
        ),
    ]



from django.apps import AppConfig


class BootstrapDarkThemeConfig(AppConfig):
    name = "allianceauth.theme.bootstrap_dark"
    label = "bootstrap_dark"
    version = "5.3.0"
    verbose_name = f"Bootstrap Dark v{version}"

    def ready(self):
        pass



import json
from pathlib import Path
from unittest import TestCase

from ..models import Guild, GuildMember, Role, User
from .factories import create_guild, create_guild_member, create_role, create_user


def _fetch_example_objects() -> dict:
    path = Path(__file__).parent / "example_objects.json"
    with path.open("r", encoding="utf-8") as fp:
        return json.load(fp)


class TestUser(TestCase):
    def test_should_create_new_object(self):
        # when
        obj = User(id="42", username=123, discriminator=456)
        # then
        self.assertEqual(obj.id, 42)
        self.assertEqual(obj.username, "123")
        self.assertTrue(obj.discriminator, "456")

    def test_should_create_from_dict(self):
        # given
        data = example_objects["users"]["80351110224678912"]
        # when
        obj = User.from_dict(data)
        # then
        self.assertEqual(obj.id, 80351110224678912)
        self.assertEqual(obj.username, "Nelly")
        self.assertEqual(obj.discriminator, "1337")


class TestRole(TestCase):
    def test_should_create_new_object_with_defaults(self):
        # when
        obj = Role(id="42", name="x" * 110)
        # then
        self.assertEqual(obj.id, 42)
        self.assertEqual(obj.name, "x" * 100)
        self.assertFalse(obj.managed)

    def test_should_create_new_object(self):
        # when
        obj = Role(id=42, name="name", managed=1)
        # then
        self.assertEqual(obj.id, 42)
        self.assertEqual(obj.name, "name")
        self.assertTrue(obj.managed)

    def test_should_create_from_dict(self):
        # given
        data = example_objects["roles"]["41771983423143936"]
        # when
        obj = Role.from_dict(data)
        # then
        self.assertEqual(obj.id, 41771983423143936)
        self.assertEqual(obj.name, "WE DEM BOYZZ!!!!!!")
        self.assertFalse(obj.managed)

    def test_should_convert_to_dict(self):
        # given
        role = create_role(id=42, name="Special Name", managed=True)
        # when/then
        self.assertDictEqual(
            role.asdict(), {"id": 42, "name": "Special Name", "managed": True}
        )

    def test_sanitize_role_name(self):
        # given
        role_name_input = "x" * 110
        role_name_expected = "x" * 100
        # when
        result = Role.sanitize_name(role_name_input)
        # then
        self.assertEqual(result, role_name_expected)


class TestGuild(TestCase):
    def test_should_create_new_object(self):
        # given
        role_a = create_role()
        # when
        obj = Guild(id="42", name=123, roles=[role_a])
        # then
        self.assertEqual(obj.id, 42)
        self.assertEqual(obj.name, "123")
        self.assertEqual(obj.roles, frozenset([role_a]))

    def test_should_create_from_dict(self):
        # given
        data = example_objects["guilds"]["2909267986263572999"]
        # when
        obj = Guild.from_dict(data)
        # then
        self.assertEqual(obj.id, 2909267986263572999)
        self.assertEqual(obj.name, "Mason's Test Server")
        (first_role,) = obj.roles
        self.assertEqual(first_role.id, 2909267986263572999)

    def test_should_raise_error_when_role_type_is_wrong(self):
        with self.assertRaises(TypeError):
            create_guild(roles=[create_role(), "invalid"])


class TestGuildMember(TestCase):
    def test_should_create_new_object(self):
        # given
        user = create_user()
        # when
        obj = GuildMember(user=user, nick="x" * 40, roles=[1, 2])
        # then
        self.assertEqual(obj.user, user)
        self.assertEqual(obj.nick, "x" * 32)
        self.assertEqual(obj.roles, frozenset([1, 2]))

    def test_should_create_from_dict_empty(self):
        # given
        data = example_objects["guildMembers"]["1"]
        # when
        obj = GuildMember.from_dict(data)
        # then
        self.assertIsNone(obj.user)
        self.assertSetEqual(obj.roles, set())
        self.assertIsNone(obj.nick)

    def test_should_create_from_dict_full(self):
        # given
        data = example_objects["guildMembers"]["2"]
        # when
        obj = GuildMember.from_dict(data)
        # then
        self.assertEqual(obj.user.username, "Nelly")
        self.assertSetEqual(obj.roles, {197150972374548480, 41771983423143936})
        self.assertEqual(obj.nick, "Nelly the great")

    def test_should_raise_error_when_user_type_is_wrong(self):
        with self.assertRaises(TypeError):
            create_guild_member(user="invalid")

    def test_should_raise_error_when_role_type_is_wrong(self):
        with self.assertRaises(TypeError):
            GuildMember(roles=[1, 2, "invalid"])

    def test_sanitize_nick(self):
        # given
        nick_input = "x" * 40
        nick_expected = "x" * 32
        # when
        result = GuildMember.sanitize_nick(nick_input)
        # then
        self.assertEqual(result, nick_expected)


example_objects = _fetch_example_objects()



from unittest import mock
from urllib import parse

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.http.response import HttpResponse
from django.test import TestCase
from django.test.client import RequestFactory
from django.urls import reverse, URLPattern

from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils


from ..decorators import decorate_url_patterns, main_character_required
from ..models import CharacterOwnership

MODULE_PATH = 'allianceauth.authentication'


class DecoratorTestCase(TestCase):
    @staticmethod
    @main_character_required
    def dummy_view(*args, **kwargs):
        return HttpResponse(status=200)

    @classmethod
    def setUpTestData(cls):
        cls.main_user = AuthUtils.create_user('main_user', disconnect_signals=True)
        cls.no_main_user = AuthUtils.create_user(
            'no_main_user', disconnect_signals=True
        )
        main_character = EveCharacter.objects.create(
            character_id=1,
            character_name='Main Character',
            corporation_id=1,
            corporation_name='Corp',
            corporation_ticker='CORP',
        )
        CharacterOwnership.objects.create(
            user=cls.main_user, character=main_character, owner_hash='1'
        )
        cls.main_user.profile.main_character = main_character

    def setUp(self):
        self.request = RequestFactory().get('/test/')

    @mock.patch(MODULE_PATH + '.decorators.messages')
    def test_login_redirect(self, m):
        setattr(self.request, 'user', AnonymousUser())
        response = self.dummy_view(self.request)
        self.assertEqual(response.status_code, 302)
        url = getattr(response, 'url', None)
        self.assertEqual(parse.urlparse(url).path, reverse(settings.LOGIN_URL))

    @mock.patch(MODULE_PATH + '.decorators.messages')
    def test_main_character_redirect(self, m):
        setattr(self.request, 'user', self.no_main_user)
        response = self.dummy_view(self.request)
        self.assertEqual(response.status_code, 302)
        url = getattr(response, 'url', None)
        self.assertEqual(url, reverse('authentication:dashboard'))

    @mock.patch(MODULE_PATH + '.decorators.messages')
    def test_successful_request(self, m):
        setattr(self.request, 'user', self.main_user)
        response = self.dummy_view(self.request)
        self.assertEqual(response.status_code, 200)


class TestDecorateUrlPatterns(TestCase):
    def test_should_add_decorator_by_default(self):
        # given
        decorator = mock.MagicMock(name="decorator")
        view = mock.MagicMock(name="view")
        path = mock.MagicMock(spec=URLPattern, name="path")
        path.callback = view
        path.lookup_str = "my_lookup_str"
        urls = [path]
        urlconf_module = urls
        # when
        decorate_url_patterns(urlconf_module, decorator)
        # then
        self.assertEqual(path.callback, decorator(view))

    def test_should_not_add_decorator_when_excluded(self):
        # given
        decorator = mock.MagicMock(name="decorator")
        view = mock.MagicMock(name="view")
        path = mock.MagicMock(spec=URLPattern, name="path")
        path.callback = view
        path.lookup_str = "my_lookup_str"
        urls = [path]
        urlconf_module = urls
        # when
        decorate_url_patterns(urlconf_module, decorator, excluded_views=["my_lookup_str"])
        # then
        self.assertEqual(path.callback, view)



import logging
from django.contrib.auth.models import Group
from django.db.models.signals import pre_save, post_save
from django.dispatch import receiver

from allianceauth.authentication.signals import state_changed

from .models import AuthGroup, ReservedGroupName

logger = logging.getLogger(__name__)


@receiver(pre_save, sender=Group)
def find_new_name_for_conflicting_groups(sender, instance, **kwargs):
    """Find new name for a group which name is already reserved."""
    new_name = instance.name
    num = 0
    while ReservedGroupName.objects.filter(name__iexact=new_name).exists():
        num += 1
        new_name = f"{instance.name}_{num}"
    instance.name = new_name


@receiver(post_save, sender=Group)
def create_auth_group(sender, instance, created, **kwargs):
    """Create the AuthGroup model when a group is created."""
    if created:
        AuthGroup.objects.create(group=instance)


@receiver(state_changed)
def check_groups_on_state_change(sender, user, state, **kwargs):
    logger.debug(
        f"Checking group memberships for {user} based on new state {state}"
    )
    state_groups = (
        user.groups.select_related("authgroup").exclude(authgroup__states=None)
    )
    for group in state_groups:
        if not group.authgroup.states.filter(id=state.id).exists():
            logger.info(
                f"Removing user {user} from group {group} due to missing state"
            )
            user.groups.remove(group)



from django.utils.timezone import now

from allianceauth.authentication.backends import StateBackend
from allianceauth.tests.auth_utils import AuthUtils

from ..discord_client.tests.factories import (
    TEST_USER_DISCRIMINATOR,
    TEST_USER_ID,
    TEST_USER_NAME,
)
from ..models import DiscordUser


def create_user(**kwargs):
    params = {"username": TEST_USER_NAME}
    params.update(kwargs)
    username = StateBackend.iterate_username(params["username"])
    user = AuthUtils.create_user(username)
    return AuthUtils.add_permission_to_user_by_name("discord.access_discord", user)


def create_discord_user(user=None, **kwargs):
    params = {
        "user": user or create_user(),
        "uid": TEST_USER_ID,
        "username": TEST_USER_NAME,
        "discriminator": TEST_USER_DISCRIMINATOR,
        "activated": now(),
    }
    params.update(kwargs)
    return DiscordUser.objects.create(**params)



from django.urls import re_path
from django.urls import path

from . import views

app_name = 'permissions_tool'

urlpatterns = [
    path('overview/', views.permissions_overview, name='overview'),
    re_path(r'^audit/(?P<app_label>[\w\-_]+)/(?P<model>[\w\-_]+)/(?P<codename>[\w\-_]+)/$', views.permissions_audit,
        name='audit'),
]



# Generated by Django 4.0.10 on 2023-05-28 15:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("authentication", "0020_userprofile_language_userprofile_night_mode"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userprofile",
            name="language",
            field=models.CharField(
                blank=True,
                choices=[
                    ("en", "English"),
                    ("de", "German"),
                    ("es", "Spanish"),
                    ("zh-hans", "Chinese Simplified"),
                    ("ru", "Russian"),
                    ("ko", "Korean"),
                    ("fr", "French"),
                    ("ja", "Japanese"),
                    ("it", "Italian"),
                    ("uk", "Ukrainian"),
                ],
                default="",
                max_length=10,
                verbose_name="Language",
            ),
        ),
    ]



from django.contrib.auth.models import User
from django.db import models


class Ips4User(models.Model):
    user = models.OneToOneField(User,
                                primary_key=True,
                                on_delete=models.CASCADE,
                                related_name='ips4')
    username = models.CharField(max_length=254)
    id = models.CharField(max_length=254)

    def __str__(self):
        return self.username

    class Meta:
        permissions = (
            ("access_ips4", "Can access the IPS4 service"),
        )



import logging

import requests

from django.contrib.auth.models import User

from allianceauth import NAME
from allianceauth.srp.providers import esi

from .models import SrpUserRequest

logger = logging.getLogger(__name__)


class SRPManager:

    @staticmethod
    def get_kill_id(killboard_link):
        num_set = '0123456789'
        kill_id = ''.join([c for c in killboard_link if c in num_set])
        return kill_id

    @staticmethod
    def get_kill_data(kill_id):
        url = ("https://zkillboard.com/api/killID/%s/" % kill_id)
        headers = {
            'User-Agent': NAME,
            'Content-Type': 'application/json',
        }
        r = requests.get(url, headers=headers)
        result = r.json()[0]
        if result:
            killmail_id = result['killmail_id']
            killmail_hash = result['zkb']['hash']
            km = esi.client.Killmails.get_killmails_killmail_id_killmail_hash(
                killmail_id=killmail_id,
                killmail_hash=killmail_hash
            ).result()
        else:
            raise ValueError("Invalid Kill ID")
        if km:
            ship_type = km['victim']['ship_type_id']
            logger.debug(
                f"Ship type for kill ID {kill_id} is {ship_type}"
            )
            ship_value = result['zkb']['totalValue']
            logger.debug(
                f"Total loss value for kill id {kill_id} is {ship_value}"
            )
            victim_id = km['victim']['character_id']
            return ship_type, ship_value, victim_id
        else:
            raise ValueError("Invalid Kill ID or Hash.")

    @staticmethod
    def pending_requests_count_for_user(user: User):
        """returns the number of open SRP requests for given user
        or None if user has no permission"""
        if user.has_perm("auth.srp_management"):
            return SrpUserRequest.objects.filter(srp_status="Pending").count()
        else:
            return None



# this module generates profile URLs for zKillboard

from urllib.parse import urljoin

from . import (
    _ESI_CATEGORY_ALLIANCE,
    _ESI_CATEGORY_CORPORATION,
    _ESI_CATEGORY_CHARACTER,
    _ESI_CATEGORY_REGION,
    _ESI_CATEGORY_SOLARSYSTEM
)


_BASE_URL = 'https://zkillboard.com'


def _build_url(category: str, eve_id: int) -> str:
    """return url to profile page for an eve entity"""

    if category == _ESI_CATEGORY_ALLIANCE:
        partial = 'alliance'

    elif category == _ESI_CATEGORY_CORPORATION:
        partial = 'corporation'

    elif category == _ESI_CATEGORY_CHARACTER:
        partial = 'character'

    elif category == _ESI_CATEGORY_REGION:
        partial = 'region'

    elif category == _ESI_CATEGORY_SOLARSYSTEM:
        partial = 'system'

    else:
        raise NotImplementedError(
            "Not implemented yet for category:" + category
        )

    url = urljoin(
        _BASE_URL,
        f'{partial}/{int(eve_id)}/'
    )
    return url


def alliance_url(eve_id: int) -> str:
    """url for page about given alliance on zKillboard"""
    return _build_url(_ESI_CATEGORY_ALLIANCE, eve_id)


def character_url(eve_id: int) -> str:
    """url for page about given character on zKillboard"""
    return _build_url(_ESI_CATEGORY_CHARACTER, eve_id)


def corporation_url(eve_id: int) -> str:
    """url for page about given corporation on zKillboard"""
    return _build_url(_ESI_CATEGORY_CORPORATION, eve_id)


def region_url(eve_id: int) -> str:
    """url for page about given region on zKillboard"""
    return _build_url(_ESI_CATEGORY_REGION, eve_id)


def solar_system_url(eve_id: int) -> str:
    return _build_url(_ESI_CATEGORY_SOLARSYSTEM, eve_id)



import logging
from typing import Optional

from requests.exceptions import HTTPError

from django.contrib.auth.models import User
from django.db import models
from django.utils.translation import gettext_lazy

from allianceauth.notifications import notify

from . import __title__
from .app_settings import DISCORD_GUILD_ID
from .core import (
    create_bot_client,
    default_bot_client,
    calculate_roles_for_user,
    user_formatted_nick
)
from .discord_client import DiscordApiBackoff
from .managers import DiscordUserManager
from .utils import LoggerAddTag


logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class DiscordUser(models.Model):
    """The Discord user account of an Auth user."""

    user = models.OneToOneField(
        User,
        primary_key=True,
        on_delete=models.CASCADE,
        related_name='discord',
        help_text='Auth user owning this Discord account'
    )
    uid = models.BigIntegerField(
        db_index=True,
        help_text='user\'s ID on Discord'
    )
    username = models.CharField(
        max_length=32,
        default='',
        blank=True,
        db_index=True,
        help_text='user\'s username on Discord'
    )
    discriminator = models.CharField(
        max_length=4,
        default='',
        blank=True,
        help_text='user\'s discriminator on Discord'
    )
    activated = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
        help_text='Date & time this service account was activated'
    )

    objects = DiscordUserManager()

    class Meta:
        permissions = (
            ("access_discord", "Can access the Discord service"),
        )

    def __str__(self):
        return f'{self.user.username} - {self.uid}'

    def __repr__(self):
        return f'{type(self).__name__}(user=\'{self.user}\', uid={self.uid})'

    def update_nickname(self, nickname: str = None) -> bool:
        """Update nickname with formatted name of main character

        Params:
        - nickname: optional nickname to be used instead of user's main

        Returns:
        - True on success
        - None if user is no longer a member of the Discord server
        - False on error or raises exception
        """
        if not nickname:
            nickname = user_formatted_nick(self.user)
        if not nickname:
            return False
        success = default_bot_client.modify_guild_member(
            guild_id=DISCORD_GUILD_ID,
            user_id=self.uid,
            nick=nickname
        )
        if success:
            logger.info('Nickname for %s has been updated', self.user)
        else:
            logger.warning('Failed to update nickname for %s', self.user)
        return success

    def update_groups(self, state_name: str = None) -> Optional[bool]:
        """update groups for a user based on his current group memberships.
        Will add or remove roles of a user as needed.

        Params:
        - state_name: optional state name to be used

        Returns:
        - True on success
        - None if user is no longer a member of the Discord server
        - False on error or raises exception
        """
        new_roles, is_changed = calculate_roles_for_user(
            user=self.user,
            client=default_bot_client,
            discord_uid=self.uid,
            state_name=state_name
        )
        if is_changed is None:
            logger.debug('User is not a member of this guild %s', self.user)
            return None
        if is_changed:
            logger.debug('Need to update roles for user %s', self.user)
            success = default_bot_client.modify_guild_member(
                guild_id=DISCORD_GUILD_ID,
                user_id=self.uid,
                role_ids=list(new_roles.ids())
            )
            if success:
                logger.info('Roles for %s have been updated', self.user)
            else:
                logger.warning('Failed to update roles for %s', self.user)
            return success
        logger.info('No need to update roles for user %s', self.user)
        return True

    def update_username(self) -> Optional[bool]:
        """Updates the username incl. the discriminator
        from the Discord server and saves it

        Returns:
        - True on success
        - None if user is no longer a member of the Discord server
        """
        member_info = default_bot_client.guild_member(
            guild_id=DISCORD_GUILD_ID, user_id=self.uid
        )
        if not member_info:
            logger.warning('%s: User not a guild member', self.user)
            return None
        self.username = member_info.user.username
        self.discriminator = member_info.user.discriminator
        self.save()
        logger.info('%s: Username has been updated', self.user)
        return True

    def delete_user(
        self,
        notify_user: bool = False,
        is_rate_limited: bool = True,
        handle_api_exceptions: bool = False
    ) -> Optional[bool]:
        """Deletes the Discount user both on the server and locally

        Params:
        - notify_user: When True will sent a notification to the user
        informing him about the deleting of his account
        - is_rate_limited: When False will disable default rate limiting (use with care)
        - handle_api_exceptions: When True method will return False
        when an API exception occurs

        Returns True when successful, otherwise False or raises exceptions
        Return None if user does no longer exist
        """
        try:
            _user = self.user
            client = create_bot_client(is_rate_limited=is_rate_limited)
            success = client.remove_guild_member(
                guild_id=DISCORD_GUILD_ID, user_id=self.uid
            )
            if success is not False:
                deleted_count, _ = self.delete()
                if deleted_count > 0:
                    if notify_user:
                        notify(
                            user=_user,
                            title=gettext_lazy('Discord Account Disabled'),
                            message=gettext_lazy(
                                'Your Discord account was disabled automatically '
                                'by Auth. If you think this was a mistake, '
                                'please contact an admin.'
                            ),
                            level='warning'
                        )
                    logger.info('Account for user %s was deleted.', _user)
                    return True
                logger.debug('Account for user %s was already deleted.', _user)
                return None

            logger.warning(
                'Failed to remove user %s from the Discord server', _user
            )
            return False

        except (HTTPError, ConnectionError, DiscordApiBackoff) as ex:
            if handle_api_exceptions:
                logger.exception(
                    'Failed to remove user %s from Discord server: %s',self.user, ex
                )
                return False
            raise ex



# Generated by Django 4.0.8 on 2023-02-05 07:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('menu', '0002_alter_menuitem_hook_function_and_more'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='menuitem',
            index=models.Index(fields=['rank'], name='menu_menuit_rank_e880ab_idx'),
        ),
    ]



# Generated by Django 2.2.8 on 2020-01-06 11:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
        ('groupmanagement', '0011_requestlog_date'),
    ]

    operations = [
        migrations.AddField(
            model_name='authgroup',
            name='group_leader_groups',
            field=models.ManyToManyField(blank=True, help_text='Group leaders can process group requests for this group specifically. Use the auth.group_management permission to allow a user to manage all groups.', related_name='leads_group_groups', to='auth.Group'),
        )
    ]



from unittest.mock import MagicMock, patch

from django.test import TestCase

from allianceauth.utils.cache import get_redis_client

MODULE_PATH = "allianceauth.utils.cache"


class RedisClientStub:
    """Substitute for a Redis client."""

    pass


class TestGetRedisClient(TestCase):
    def test_should_work_with_aa2_api(self):
        # given
        mock_django_cache = MagicMock()
        mock_django_cache.get_master_client.return_value = RedisClientStub()
        # when
        with patch(MODULE_PATH + ".django_redis", None), patch.dict(
            MODULE_PATH + ".caches", {"default": mock_django_cache}
        ):
            client = get_redis_client()
        # then
        self.assertIsInstance(client, RedisClientStub)

    def test_should_work_with_aa3_api(self):
        # given
        mock_django_redis = MagicMock()
        mock_django_redis.get_redis_connection.return_value = RedisClientStub()
        # when
        with patch(MODULE_PATH + ".django_redis", mock_django_redis), patch.dict(
            MODULE_PATH + ".caches", {"default": None}
        ):
            client = get_redis_client()
        # then
        self.assertIsInstance(client, RedisClientStub)



# Generated by Django 1.10.1 on 2016-09-09 23:19

from django.db import migrations


def create_permission(apps, schema_editor):
    User = apps.get_model('auth', 'User')
    ContentType = apps.get_model('contenttypes', 'ContentType')
    Permission = apps.get_model('auth', 'Permission')
    ct = ContentType.objects.get_for_model(User)
    Permission.objects.get_or_create(codename="view_fleetup", content_type=ct, name="view_fleetup")


class Migration(migrations.Migration):

    dependencies = [
        ('authentication', '0013_service_modules'),
    ]

    operations = [
        migrations.RunPython(create_permission, migrations.RunPython.noop)
    ]
